import os
from krixik.pipeline_builder.utilities.chain_checker import chain_check
from krixik.utilities.validators.data.utilities.read_config import check_inverse_config
from krixik.modules.utilities.io_validator import is_valid


def input_check(local_file_path: str, module_chain: list) -> None:
    if not isinstance(local_file_path, str):
        raise TypeError(f"local_file_path must be a string: {local_file_path}")

    if not os.path.exists(local_file_path):
        raise FileExistsError(f"FAILURE: local_file_path does not exist - {local_file_path}")
    chain_check(module_chain)

    first_module = module_chain[0]
    first_module_input_format = first_module.input_format
    file_ext = "." + local_file_path.split(".")[-1]
    file_ext_format = check_inverse_config(file_ext)
    if file_ext_format != first_module_input_format:
        raise TypeError(f"file extension '{file_ext}' does not match the expected input format {first_module_input_format}")
    is_valid(first_module.name, local_file_path)



import requests
from typing import Optional
import json
import os
from krixik.system_builder.functions import update_endpoint


def update(
    self,
    *,
    file_id: str,
    file_name: Optional[str],
    symbolic_directory_path: Optional[str],
    symbolic_file_path: Optional[str],
    file_tags: Optional[dict],
    file_description: Optional[str],
    expire_time: Optional[str],
    verbose: bool = True,
) -> dict:
    # check that file_id is not empty
    if file_id is None:
        raise ValueError("invalid file_id - file_id cannot be None")

    # check query arguments
    # check that file descriptors are not all empty
    if (
        file_name is None
        and symbolic_directory_path is None
        and symbolic_file_path is None
        and file_tags is None
        and file_description is None
        and expire_time is None
    ):
        raise ValueError(
            "at least one of the following update arguments must be given: file_name, symbolic_directory_path, symbolic_file_path, file_tags, file_description, expire_time"
        )

    # check that either file_name and symbolic_directory_path are not empty or that symbolic_file_path is not empty
    if (file_name is not None or symbolic_directory_path is not None) and symbolic_file_path is not None:
        raise ValueError("file_name and symbolic_directory_path cannot both be given if symbolic_file_path is given")

    # if symbolic_file_path is not empty, split symbolic_file_path into symbolic_directory_path and file_name
    if symbolic_file_path is not None:
        symbolic_directory_path = os.path.dirname(symbolic_file_path)
        file_name = os.path.basename(symbolic_file_path)

    if hasattr(self, "_KrixikBasePipeline__pipeline"):
        pipeline = self._KrixikBasePipeline__pipeline
    elif hasattr(self, "_KrixikSearchPipeline__pipeline"):
        pipeline = self._KrixikSearchPipeline__pipeline
    else:
        raise ValueError("pipeline not found in self")

    if hasattr(self, "_KrixikBasePipeline__version"):
        version = self._KrixikBasePipeline__version
    elif hasattr(self, "_KrixikSearchPipeline__version"):
        version = self._KrixikSearchPipeline__version
    else:
        raise ValueError("version not found in self")

    # prep payload data
    payload_data = {
        "pipeline": pipeline,
        "version": version,
        "file_id": file_id,
        "file_name": file_name,
        "symbolic_directory_path": symbolic_directory_path,
        "file_tags": file_tags,
        "file_description": file_description,
        "expire_time": expire_time,
    }

    if hasattr(self, "_KrixikBasePipeline__api_key") and hasattr(self, "_KrixikBasePipeline__api_url"):
        api_key = self._KrixikBasePipeline__api_key
        api_url = self._KrixikBasePipeline__api_url
    elif hasattr(self, "_KrixikSearchPipeline__api_key") and hasattr(self, "_KrixikSearchPipeline__api_url"):
        api_key = self._KrixikSearchPipeline__api_key
        api_url = self._KrixikSearchPipeline__api_url
    else:
        raise ValueError("api_key and api_url not found in self")

    # prep headers
    headers = {"Content-Type": "text/plain", "krixikApiKey": api_key}

    try:
        # make request
        response = requests.post(
            api_url + update_endpoint,
            headers=headers,
            json=payload_data,
            timeout=60,
        )

        # return response
        results = json.loads(response.text)
        self.request_id = results["request_id"]
        status_code_dict = {"status_code": response.status_code}
        status_code_dict.update(results)
        return status_code_dict
    except Exception as e:
        raise ValueError(f"update failed with request exception {e}")
    finally:
        self._reset_class_variables()



import copy
import re
from krixik.utilities.utilities import vprint
from krixik.utilities.validators.system.base.file_names import (
    individual_file_name_checker,
)
from krixik.utilities.validators.system import (
    SYMBOLIC_DIRECTORY_PATH_MAX_LENGTH,
    MAX_SYMBOLIC_PATH_COUNT,
)


def lower_case_symbolic_paths(symbolic_paths: list) -> list:
    return [s.lower() for s in symbolic_paths]


# symbolic directory paths
# define pattern for valid directory path
symbolic_directory_path_valid_pattern = r"^/[a-zA-Z0-9 _-]+(?:/[a-zA-Z0-9 _-]+)*(?:/$)?$"


def is_valid_directory_path(path: str) -> bool:
    return bool(re.match(symbolic_directory_path_valid_pattern, path))


# check individual symbolic directory path
def individual_symbolic_directory_path_checker(path: str, phase: str = "other") -> None:
    if path is not None:
        # check that path is a string
        if not isinstance(path, str):
            raise TypeError(f"invalid symbolic_directory_path: not a string - {path}")

        # check that path length is greater than 0
        if len(path) == 0:
            raise ValueError(f"invalid symbolic_directory_path: length is 0 - {path}")

        # check that path length is less than SYMBOLIC_DIRECTORY_PATH_MAX_LENGTH
        if len(path) > SYMBOLIC_DIRECTORY_PATH_MAX_LENGTH:
            raise ValueError(f"invalid symbolic_directory_path: length is greater than 128 (current maximum length allowable) - {path}")

        # check if path ends with /
        if len(path) > 1 and path[-1] == "/":
            raise ValueError(f"invalid symbolic_directory_path: cannot end with / - {path}")

        # check if path starts with /
        if path[0] != "/":
            raise ValueError(f"invalid symbolic_directory_path: must start with / - {path}")

        # type check that path - minus possible * at the end - is a valid directory path
        test_path = copy.deepcopy(path)
        if phase != "process":
            if path[-1] == "*":
                test_path = path[:-1]
            if test_path != "/":
                res = is_valid_directory_path(test_path)
                if res is False:
                    raise ValueError(
                        f"invalid symbolic_directory_path: please check that your symbolic path takes a unix form: {symbolic_directory_path_valid_pattern} \n You can only use alphanumeric characters, spaces, underscores, dashes, and forward slashes. \n Example symbolic_diretory_path: /home/files/invoices"
                    )
        else:
            if path != "/":
                res = is_valid_directory_path(test_path)
                if res is False:
                    raise ValueError(
                        f"invalid symbolic_directory_path: please check that your symbolic path takes a unix form: {symbolic_directory_path_valid_pattern} \n You can only use alphanumeric characters, spaces, underscores, dashes, and forward slashes. \n Example symbolic_diretory_path: /home/files/invoices"
                    )
            # else:
            #     raise ValueError(
            #         f"invalid symbolic_directory_path: cannot be / - {path}"
            #     )

        # check if path contains /etc and more than two instances of '/'
        if "/etc" in path and path != "/etc/*" and path.count("/") >= 2:
            raise ValueError(f"Invalid symbolic_directory_path: cannot contain /etc - {path}")


# check symbolic directory paths
def symbolic_directory_paths_checker(symbolic_directory_paths: list, phase="other", verbose: bool = True) -> None:
    if symbolic_directory_paths is not None:
        # check that symbolic_directory_paths is a list
        if not isinstance(symbolic_directory_paths, list):
            raise TypeError("symbolic_directory_paths is not a list")

        if len(symbolic_directory_paths) == 0:
            raise ValueError("symbolic_directory_paths is empty")

        if len(symbolic_directory_paths) > MAX_SYMBOLIC_PATH_COUNT:
            raise ValueError(f"symbolic_directory_paths contains more than {MAX_SYMBOLIC_PATH_COUNT} paths - {symbolic_directory_paths}")

        # check that symbolic_directory_paths is not empty
        if symbolic_directory_paths is not None:
            # check that all paths in symbolic_directory_paths are valid paths
            for path in symbolic_directory_paths:
                # check individual path
                individual_symbolic_directory_path_checker(path, phase=phase)

            # check for duplicates
            if len(symbolic_directory_paths) != len(set(symbolic_directory_paths)):
                # find duplicates
                duplicates = []
                for path in symbolic_directory_paths:
                    if symbolic_directory_paths.count(path) > 1:
                        duplicates.append(path)

                # print error message
                vprint(
                    f"WARNING: symbolic_directory_paths contains duplicate entries - {duplicates}",
                    verbose=verbose,
                )
                vprint(
                    "INFO: You can remove duplicates by using Python's list(set()) function on your list of symbolic_directory_paths",
                    verbose=verbose,
                )
        else:
            raise ValueError("symbolic_directory_paths is None")


# symbolic file paths
def individual_symbolic_file_path_checker(file_path: str, valid_extensions: list) -> None:
    if file_path is not None:
        # check that file_path is a string
        if not isinstance(file_path, str):
            raise TypeError(f"invalid symbolic_file_path: not a string - {file_path}")

        # check that file_path has length greater than 0
        if len(file_path) == 0:
            raise ValueError("Input symbolic_file_path has length 0 - a valid file path must at least be forward slash '/' indicating root directory")

        # check that file_path starts with '/'
        if file_path[0] != "/":
            raise ValueError(f"Input symbolic_file_path did not begin with a forward slash ('/') - {file_path}")

        # split file_path into directory_path and file_name
        try:
            split = file_path.split("/")
            directory_path = "/".join(split[:-1]) if split[:-1] != [""] else "/"
            file_name = split[-1]
        except Exception as e:
            raise ValueError(f"invalid symbolic_file_path: unable to split into directory_path and file_name - {file_path} - {e}")

        try:
            # check that file_name is a valid file name
            individual_file_name_checker(file_name, valid_extensions)

            # check that directory_path is a valid directory path
            individual_symbolic_directory_path_checker(directory_path)
        except Exception as e:
            if "symbolic_directory_path" in str(e):
                raise ValueError(
                    f"the symbolic_directory_path portion - {directory_path} -  of your symbolic_file_path - {file_path} - failed with exception: {e} "
                )
            if "file_name" in str(e):
                if file_name == "*":
                    raise ValueError(
                        f"the file_name portion of your symbolic_file_path - {file_path} is an asterisk '*' which is not allowed - you can only use an asterisk at the end of a symbolic_directory_path"
                    )
                else:
                    raise ValueError(f"the file_name portion - {file_name} - of your symbolic_file_path - {file_path} - failed with exception: {e} ")


def symbolic_file_paths_checker(symbolic_file_paths: list, valid_extensions: list, verbose: bool = True) -> None:
    if symbolic_file_paths is not None:
        # check that symbolic_directory_paths is a list
        if not isinstance(symbolic_file_paths, list):
            raise TypeError("symbolic_file_paths is not a list")

        # check that symbolic_file_paths is not empty
        if len(symbolic_file_paths) == 0:
            raise ValueError("symbolic_file_paths is empty")

        if len(symbolic_file_paths) > MAX_SYMBOLIC_PATH_COUNT:
            raise ValueError(f"symbolic_file_paths contains more than {MAX_SYMBOLIC_PATH_COUNT} paths - {symbolic_file_paths}")

        # check that symbolic_directory_paths is not empty
        if symbolic_file_paths is not None:
            # check that all paths in symbolic_directory_paths are valid paths
            for path in symbolic_file_paths:
                # check individual path
                individual_symbolic_file_path_checker(path, valid_extensions)

        # check for duplicates
        if len(symbolic_file_paths) != len(set(symbolic_file_paths)):
            # find duplicates
            duplicates = []
            for path in symbolic_file_paths:
                if symbolic_file_paths.count(path) > 1:
                    duplicates.append(path)

            # print error message
            vprint(
                f"WARNING: symbolic_file_paths contains duplicate entries - {duplicates}",
                verbose=verbose,
            )
            vprint(
                "INFO: You can remove duplicates by using Python's list(set()) function on your list of symbolic_file_paths",
                verbose=verbose,
            )



import time
import copy
import os
import tempfile
from typing import Optional, Tuple
from krixik.__version__ import __version__
from krixik.utilities.validators.system import EXPIRE_TIME_DEFAULT
from krixik.system_builder.functions.checkin import checkin
from krixik.system_builder.functions.list_files import list_files
from krixik.system_builder.functions.fetch_output import fetch_output
from krixik.system_builder.functions.show_tree import show_post
from krixik.system_builder.functions.show_tree import show_illustration
from krixik.system_builder.functions.update import update
from krixik.system_builder.functions.process import process_local_file
from krixik.system_builder.functions.process import get_presigned_url
from krixik.system_builder.functions.delete import delete_server_files
from krixik.system_builder.functions.check_process_status import check_process_status
from krixik.system_builder.functions.check_process_status import process_status_reporter
from krixik.utilities.utilities import vprint
from krixik.utilities.file_name_generator import generate_random_file_name
from krixik.system_builder.functions.checkin import check_init_decorator
from krixik.utilities.converters.utilities.decorators import datatype_converter_wrapper
from krixik.utilities.validators.utilities.decorators import type_check_inputs
from krixik.modules.utilities.decorators import type_check_inputs as type_check_modules
from krixik.system_builder.utilities.decorators import kwargs_checker


class KrixikBasePipeline:
    def __init__(
        self,
        *,
        pipeline: Optional[str] = None,
        module_chain: list,
        output_process_keys: list,
        api_key: Optional[str] = None,
        api_url: Optional[str] = None,
        api_check_val: Optional[int] = None,
    ) -> None:
        # inherit necessary custom pipeline variables
        self.__pipeline = pipeline
        self.__pipeline_ordered_modules = module_chain
        self.__pipeline_output_process_keys = output_process_keys
        self.__first_module = self.__pipeline_ordered_modules[0]
        self.__last_module = self.__pipeline_ordered_modules[-1]

        self.__version = __version__
        self.__api_key = api_key
        self.__api_url = api_url
        self.__api_check_val = api_check_val
        self.converted_file_path = None
        self.cleaned_file_path = None
        self.file_id = None
        self.request_id = None
        self.process_id = None
        self.upload_results = None
        self.wait_for_process = False
        self.file_name = None
        self.symbolic_directory_path = None
        self.file_tags = None
        self.file_description = None
        self.modules = None
        self.local_file_path = None
        self.__local_conversion_directory = tempfile.gettempdir()

    @property
    def module_chain(self):
        return self.__pipeline_ordered_modules

    @property
    def pipeline_output_process_keys(self):
        return self.__pipeline_output_process_keys

    @property
    def first_module(self):
        return self.__first_module

    @property
    def last_module(self):
        return self.__last_module

    @property
    def local_conversion_directory(self):
        return self.__local_conversion_directory

    @property
    def api_key(self):
        return self.__api_key

    @property
    def api_url(self):
        return self.__api_url

    @property
    def api_check_val(self):
        return self.__api_check_val

    @property
    def version(self):
        return self.__version

    @property
    def pipeline(self):
        return self.__pipeline

    @pipeline.setter
    def pipeline_name(self, value):
        self.__pipeline = value

    @type_check_inputs
    def init(self, *, api_key: Optional[str], api_url: Optional[str]) -> None:
        """initialize pipeline with an api_key and api_url

        Parameters
        ----------
        api_key : str | None
            valid api_key, by default None
        api_url : str | None
            valid api_url, by default None
        """
        self.__api_key = api_key
        self.__api_url = api_url
        self.__api_check_val = checkin(self.__api_key, self.__api_url)

    @check_init_decorator
    def __delete_server_files(self) -> dict:
        return delete_server_files(self)

    @kwargs_checker
    @check_init_decorator
    @type_check_inputs
    def delete(self, *, file_id: str = None) -> dict:
        """delete a file from the server via its file_id for a given pipeline

        Parameters
        ----------
        file_id : str
            the file_id of the file to be deleted, by default None

        Returns
        -------
        dict
            a dictionary containing the status_code, request_id, and message indicating success or failure
        """
        if file_id is None:
            raise ValueError("please provide a file_id")
        self.file_id = file_id
        return self.__delete_server_files()

    @kwargs_checker
    @check_init_decorator
    @type_check_inputs
    def list(
        self,
        *,
        file_ids: Optional[list] = None,
        file_names: Optional[list] = None,
        symbolic_directory_paths: Optional[list] = None,
        symbolic_file_paths: Optional[list] = None,
        file_tags: Optional[list] = None,
        sort_order: str = "descending",
        max_files: Optional[int] = None,
        created_at_start: Optional[str] = None,
        created_at_end: Optional[str] = None,
        last_updated_start: Optional[str] = None,
        last_updated_end: Optional[str] = None,
        verbose: bool = False,
    ) -> dict:
        """list files on the server for a given pipeline

        Parameters
        ----------
        file_ids : list | None, optional
            list of file_ids to filter by, by default None
        file_names : list | None, optional
            list of file_names to filter by, by default None
        symbolic_directory_paths : list | None, optional
            list of symbolic_directory_paths to filter by, by default None
        symbolic_file_paths : list | None, optional
            list of symbolic_file_paths to filter by, by default None
        file_tags : list | None, optional
            list of file_tags to filter by, by default None
        sort_order : str, optional
            the sort order of the results, by default "descending"
        max_files : int | None, optional
            the maximum number of files to return, by default None
        created_at_start : str | None, optional
            start date for created_at filter, by default None
        created_at_end : str | None, optional
            end date for created_at filter, by default None
        last_updated_start : str | None, optional
            start date for last_updated filter, by default None
        last_updated_end : str | None, optional
            end date for last_updated filter, by default None
        verbose : bool, optional
            whether to print verbose output, by default True

        Returns
        -------
        dict
            a dictionary containing the status_code, request_id, any return items, and message indicating success or failure
        """
        if sort_order == "global":
            raise ValueError("sort order must be either 'descending' or 'ascending'")
        return list_files(
            self,
            file_ids=file_ids,
            file_names=file_names,
            symbolic_directory_paths=symbolic_directory_paths,
            symbolic_file_paths=symbolic_file_paths,
            file_tags=file_tags,
            sort_order=sort_order,
            max_files=max_files,
            created_at_start=created_at_start,
            created_at_end=created_at_end,
            last_updated_start=last_updated_start,
            last_updated_end=last_updated_end,
            verbose=verbose,
        )

    @kwargs_checker
    @check_init_decorator
    @type_check_inputs
    def fetch_output(self, *, file_id: str, local_save_directory: str = os.path.abspath("")) -> Optional[dict]:
        """fetch the output of a file from the server via its file_id for a given pipeline

        Parameters
        ----------
        file_id : str
            the file_id of the file to be fetched, by default None
        local_save_directory: str, optional
            the local save directory to store process output, default is current working directory

        Returns
        -------
        dict | None
            a dictionary containing the status_code, request_id, and message indicating success or failure
        """
        return fetch_output(self, file_id, local_save_directory)

    @kwargs_checker
    @check_init_decorator
    @type_check_inputs
    def update(
        self,
        *,
        file_id: str = None,
        file_name: Optional[str] = None,
        symbolic_directory_path: Optional[str] = None,
        symbolic_file_path: Optional[str] = None,
        file_tags: Optional[dict] = None,
        file_description: Optional[str] = None,
        expire_time: Optional[str] = None,
        verbose: bool = True,
    ) -> dict:
        """update the metadata of a file def show_on the server via its file_id for a given pipeline

        Parameters
        ----------
        file_id : str, optional
            the file_id of the file to be updated, by default None
        file_name : str, optional
            the new file_name of the file, by default None
        symbolic_directory_path : str, optional
            the new symbolic_directory_path of the file, by default None
        symbolic_file_path : str, optional
            the new symbolic_file_path of the file, by default None
        file_tags : dict, optional
            the new file_tags of the file, by default None
        file_description : str, optional
            the new file_description of the file, by default None
        expire_time : str, optional
            the new expire_time of the file, by default None
        verbose : bool, optional
            whether to print verbose output, by default True

        Returns
        -------
        dict
            a dictionary containing the status_code, request_id, and message indicating success or failure
        """
        return update(
            self,
            file_id=file_id,
            file_name=file_name,
            symbolic_directory_path=symbolic_directory_path,
            symbolic_file_path=symbolic_file_path,
            file_tags=file_tags,
            file_description=file_description,
            expire_time=expire_time,
            verbose=verbose,
        )

    def _get_presigned_url(self, payload_data: dict) -> Tuple[bool, dict]:
        return get_presigned_url(self, payload_data)

    def _reset_class_variables(self) -> None:
        self.file_name = None
        self.symbolic_directory_path = None
        self.file_tags = None
        self.file_description = None
        self.modules = None
        self.local_file_path = None
        self.__presigned_post_url_results = None
        self.file_id = None
        self.request_id = None
        self.process_id = None
        self.upload_results = None

    def __upload_file_to_s3_via_presignedurl(self, verbose: bool = True) -> None:
        # process local file to s3 via presigned url
        upload_status_repeater_success, upload_response = process_local_file(self)

        try:
            if upload_response.status_code == 204:
                if not self.wait_for_process:
                    vprint("INFO: wait_for_process is set to False.", verbose=verbose)
                    vprint(
                        "INFO: File upload complete, processing has now started on our servers.  ide control will now be returned to you.",
                        verbose=verbose,
                    )
                    vprint(
                        "INFO: You will be able to check the status of this file's processing using the .process_status api as shown below.",
                        verbose=verbose,
                    )
                    vprint(
                        f"INFO: This process's request_id is: {self.process_id}.",
                        verbose=verbose,
                    )
                    vprint(
                        f"INFO: Check the status of this file's processing: .process_status(request_id='{self.process_id}').",
                        verbose=verbose,
                    )
                    return

                if self.wait_for_process:
                    vprint("INFO: File process and processing status:", verbose=verbose)
                    # check process status every timestep seconds for maximum of 600 seconds
                    timestep = 3
                    max_count = 200

                    # check process status for the first time
                    file_id, process_status, failure_status, message = check_process_status(self, process_id=self.process_id)

                    if failure_status is not None:
                        self.__delete_server_files()
                        if "failure_module" in list(failure_status.keys()):
                            raise ValueError(
                                f"processes associated with file_id {file_id} and request_id {self.process_id} failed at module {failure_status['failure_module']}"
                            )
                        else:
                            raise ValueError(f"processes associated with file_id {file_id} and request_id {self.process_id} failed")

                    prev_process_status = None
                    process_status_reporter(self, prev_process_status, process_status, verbose=verbose)
                    process_complete = False
                    process_count = 0
                    while not process_complete and process_count < max_count - 1:
                        # create process_status memory dict
                        prev_process_status = copy.deepcopy(process_status)

                        # check status
                        file_id, process_status, failure_status, message = check_process_status(self, process_id=self.process_id)

                        if failure_status is not None:
                            self.__delete_server_files()
                            if "failure_module" in list(failure_status.keys()):
                                raise ValueError(
                                    f"processes associated with request_id '{failure_status['process_id']}' failed at module '{failure_status['failure_module']}'"
                                )
                            else:
                                raise ValueError(f"processes associated with request_id '{failure_status['process_id']}' failed")

                        # report status
                        process_complete = process_status_reporter(self, prev_process_status, process_status, verbose=verbose)

                        process_count += 1
                        time.sleep(timestep)

                    if not process_complete or process_count == max_count - 1:
                        exception_message = f"file process status check timed out - the request_id of this failed process is {self.process_id}."
                        self.__delete_server_files()

                        raise ValueError(exception_message)
                    vprint("SUCCESS: pipeline process complete.", verbose=verbose)
            else:
                exception_message = f"pipeline process failed with status code: {upload_response.status_code} - the request_id of this failed process is {self.process_id}"
                self.__delete_server_files()

                raise ValueError(exception_message)
        except ValueError as err:
            raise err

    @kwargs_checker
    @check_init_decorator
    @datatype_converter_wrapper
    @type_check_modules
    @type_check_inputs
    def process(
        self,
        *,
        file_name: Optional[str] = None,
        symbolic_directory_path: Optional[str] = None,
        symbolic_file_path: Optional[str] = None,
        local_file_path: Optional[str] = None,
        file_tags: Optional[list] = None,
        file_description: Optional[str] = None,
        modules: Optional[dict] = {},
        expire_time: Optional[int] = None,
        verbose: bool = True,
        wait_for_process: bool = True,
        local_save_directory: str = os.path.abspath(""),
        download_output: bool = True,
        og_local_file_path: Optional[str] = None,
    ) -> Optional[dict]:
        """process a file to the server for a given pipeline

        Parameters
        ----------
        file_name : str, optional
            the name of the file, by default None
        symbolic_directory_path : str, optional
            the symbolic directory path of the file, by default None
        symbolic_file_path : str, optional
            the symbolic file path of the file, by default None
        local_file_path : str, optional
            the local file path of the file, by default None
        file_tags : list, optional
            the tags of the file, by default None
        file_description : str, optional
            the description of the file, by default None
        local_save_directory : str, optional
            the local save directory of the file, by default None
        modules : dict, optional
            the modules selected to process associated input file, by default None
        expire_time : int, optional
            the expire time of the file, by default None
        verbose : bool, optional
            whether to print verbose output, by default True
        wait_for_process : bool, optional
            whether to process the file asynchronously, by default False
        local_save_directory: str
            local directory for process output, by default os.getcwd()
        download_output: bool
            boolean switch, download process output (set True) or not (set False)
        og_local_file_path: str, optional
            local file path used for any file conversion - controlled internally
        Returns
        -------
        dict | None
            a dictionary containing the status_code, request_id, and message indicating success or failure
        """

        if local_file_path is None:
            raise ValueError("local_file_path cannot be empty")

        if (file_name is not None or symbolic_directory_path is not None) and symbolic_file_path is not None:
            raise ValueError("file_name and symbolic_directory_path cannot both be given if symbolic_file_path is given")

        if symbolic_directory_path is None:
            symbolic_directory_path = "/etc"
            vprint(
                "INFO: symbolic_directory_path was not set by user - setting to default of /etc",
                verbose=verbose,
            )

        if file_name is None:
            file_name = generate_random_file_name(local_file_path.split(".")[-1])
            vprint(
                f"INFO: file_name was not set by user - setting to random file name: {file_name}",
                verbose=verbose,
            )

        if expire_time is None:
            vprint(
                f"INFO: expire_time was not set by user - setting to default of {EXPIRE_TIME_DEFAULT} seconds",
                verbose=verbose,
            )
            expire_time = EXPIRE_TIME_DEFAULT

        vprint(f"INFO: wait_for_process is set to {wait_for_process}.", verbose=verbose)

        # report date of expire_time based on current_time and expire_time
        current_time = time.time()
        expire_time_date = str(time.ctime(current_time + int(expire_time))).strip()
        vprint(
            f"INFO: file will expire and be removed from you account in {expire_time} seconds, at {expire_time_date} UTC",
            verbose=verbose,
        )

        self.local_file_path = local_file_path
        self.modules = modules
        self.wait_for_process = wait_for_process
        self.file_name = file_name
        self.symbolic_directory_path = symbolic_directory_path
        self.file_tags = file_tags
        self.file_description = file_description

        payload_data = {
            "pipeline": self.__pipeline,
            "pipeline_ordered_modules": self.__pipeline_ordered_modules,
            "pipeline_output_process_keys": self.__pipeline_output_process_keys,
            "modules": self.modules,
            "version": self.__version,
            "file_name": self.file_name,
            "symbolic_directory_path": (self.symbolic_directory_path if self.symbolic_directory_path is not None else "/etc"),
            "file_tags": self.file_tags if self.file_tags is not None else [],
            "file_description": (self.file_description if self.file_description is not None else ""),
            "expire_time": expire_time,
        }

        # get presigned url and file_id
        upload_check, self.upload_results = self._get_presigned_url(payload_data=payload_data)

        if not upload_check:
            # return failure signal
            raise ValueError(self.upload_results["message"])

        vprint(
            f"INFO: {self.__pipeline} file process and input processing started...",
            verbose=verbose,
        )
        if "warnings" in self.upload_results:
            for warning in self.upload_results["warnings"]:
                vprint(warning, verbose=verbose)
            vprint("INFO: metadata can be updated using the .update api.", verbose=verbose)

        # unpack upload_results
        self.__presigned_post_url_results = self.upload_results["presigned_post_url_results"]

        self.file_id = self.upload_results["file_id"]
        self.process_id = self.upload_results["request_id"]
        vprint(
            f"INFO: This process's request_id is: {self.process_id}",
            verbose=verbose,
        )

        # package output data
        output_data = {
            "file_id": self.file_id,
            "request_id": self.process_id,
            "file_name": self.file_name,
            "symbolic_directory_path": self.symbolic_directory_path,
            "file_tags": self.file_tags,
            "file_description": self.file_description,
        }

        # process s3 file via presigned url
        self.__upload_file_to_s3_via_presignedurl(verbose=verbose)

        # reset class variables
        self._reset_class_variables()

        if not self.wait_for_process:
            return output_data

        if not download_output:
            vprint("INFO: downlaod_output set to False, not fetching output", verbose=verbose)

        if download_output and output_data is not None:
            try:
                file_id = output_data["file_id"]
                file_output = self.fetch_output(file_id=file_id, local_save_directory=local_save_directory)
                vprint("SUCCESS: process output downloaded.", verbose=verbose)
                return file_output
            except Exception as e:
                raise e
        output_data["status_code"] = 200
        return output_data

    @kwargs_checker
    @check_init_decorator
    @type_check_inputs
    def show_tree(
        self,
        *,
        symbolic_directory_path: str = "/*",
        max_files: int = 1000,
        verbose: bool = True,
    ) -> Optional[dict]:
        """show the tree of files on the server for a given pipeline

        Parameters
        ----------
        symbolic_directory_path : str,
            symbolic_directory_path or stump to visualize, by default /*
        max_files : int
            the maximum number of files to return, by default 1000
        verbose : bool
            whether to print verbose output, by default False

        Returns
        -------
        dict | None
            a dictionary containing the status_code, request_id, and message indicating success or failure
        """

        # make request
        results = show_post(self, symbolic_directory_path=symbolic_directory_path, max_files=max_files, verbose=verbose)

        # show illustration
        show_illustration(self, results=results)

        return_object = {
            "status_code": results["status_code"],
            "items": results["items"],
            "request_id": results["request_id"],
            "message": results["message"],
            "warnings": results["warnings"],
        }

        self._reset_class_variables()
        return return_object

    @kwargs_checker
    @check_init_decorator
    @type_check_inputs
    def process_status(self, request_id: str = None) -> dict:
        """check the status of a file processing for a given pipeline

        Parameters
        ----------
        request_id : str
            the request_id of the process to be checked, by default None

        Returns
        -------
        dict | None
            a string or dictionary containing the status_code, request_id, and message indicating success or failure
        """

        if request_id is None:
            raise TypeError("process request_id cannot be empty")

        # check status
        file_id, process_status, failure_status, message = check_process_status(self, process_id=request_id)

        report = {}
        report["status_code"] = 200
        report["request_id"] = self.request_id
        report["file_id"] = file_id
        report["message"] = message
        report["pipeline"] = self.__pipeline

        if process_status is not None:
            report["process_status"] = process_status
            report["overall_status"] = "complete"
            cleaned_process_status = {}
            keys = list(process_status.keys())
            for key in keys:
                cleaned_process_status[key.split(".")[0]] = process_status[key]
                if process_status[key] is not True:
                    report["overall_status"] = "ongoing"
            report["process_status"] = cleaned_process_status
            return report
        if failure_status is not None:
            report["failure_status"] = failure_status
            report["overall_status"] = "failed"
            return report
        return report

    def __getattr__(self, attr):
        if attr == "keyword_search":
            raise AttributeError(
                f"your pipeline has no attribute '{attr}' because its module_chain does not end with keyword-db: - {self.module_chain}"
            )
        if attr == "semantic_search":
            raise AttributeError(
                f"your pipeline has no attribute '{attr}' because its module_chain does not end with vector-db: - {self.module_chain}"
            )
        raise AttributeError(f"pipelines do not have the attribute '{attr}'")



import json
import copy
import importlib
from krixik.utilities.validators.data import available_data_types


def format_checker(format_: str, local_file_path: str) -> None:
    if format_ not in available_data_types:
        raise ValueError(f"FAILURE: format {format_} not yet supported")

    if format_ == "json":
        try:
            with open(local_file_path, "r") as file:
                data = json.load(file)
                first_dict = data[0]
        except Exception as e:
            raise ValueError(f"FAILURE: loading json file: {e}")


def process_key_checker(required_structure: dict, file_path: str) -> None:
    if "process_key" not in required_structure.keys():
        raise ValueError(f"FAILURE: process_key not found in required_structure")
    with open(file_path, "r") as file:
        data = json.load(file)
    data_copy = copy.deepcopy(data)
    if isinstance(data, list):
        data_copy = data_copy[0]
    else:
        raise ValueError("FAILURE: input data must be a list of dictionaries")
    if not isinstance(data_copy, dict):
        raise ValueError("FAILURE: input data must be a list of dictionaries")
    process_key = required_structure["process_key"]
    if process_key not in data_copy.keys():
        raise ValueError(f"FAILURE: required process_key {process_key} not found in data")


def validate_json_file_structure(required_structure: dict, file_path: str) -> None:
    try:
        # test that format is correct
        format_checker(required_structure["format"], file_path)

        # check that input file_path contains process_key
        if required_structure["format"] == "json":
            process_key_checker(required_structure, file_path)

    except Exception as e:
        raise e


def is_valid(module_name: str, file_path: str) -> None:
    try:
        module_io_path = f"krixik.modules.{module_name}.io"
        module = importlib.import_module(module_io_path)
        InputStructure = module.InputStructure
        input_structure = InputStructure().__dict__

        validate_json_file_structure(input_structure, file_path)
    except Exception as e:
        raise e



import json
from typing import Dict
import requests
from krixik.system_builder.functions import cap_endpoint


def cap_check(api_key: str, api_url: str, version: str) -> dict:
    if api_key is None or api_url is None:
        raise ValueError("api_key and api_url cannot be None")

    try:
        # prep headers
        headers: Dict[str, str] = {"Content-Type": "text/plain", "krixikApiKey": api_key}
        payload_data = {"version": version, "pipeline": None, "pipeline_ordered_modules": []}

        # make post
        response = requests.post(
            (api_url or "") + cap_endpoint,
            headers=headers,
            json=payload_data,
            timeout=60,
        )
        results = json.loads(response.text)
        status_code_dict = {"status_code": response.status_code}
        status_code_dict.update(results)
        return status_code_dict
    except requests.exceptions.HTTPError as e:
        raise requests.exceptions.HTTPError(f"cap check failed with HTTPError {e}")
    except requests.exceptions.ConnectionError as e:
        raise requests.exceptions.ConnectionError(f"cap check failed with ConnectionError {e}")
    except requests.exceptions.Timeout as e:
        raise requests.exceptions.Timeout(f"cap check failed with Timeout {e}")
    except requests.exceptions.RequestException as e:
        raise requests.exceptions.RequestException(f"cap check failed with request exception {e}")
    except Exception as e:
        raise ValueError(f"cap check failed with request exception {e}")



from typing import Optional


def download_output_checker(download_output: Optional[bool] = None) -> None:
    if download_output is None:
        raise TypeError("download_output must be a boolean")
    if not isinstance(download_output, bool):
        raise TypeError("download_output must be a boolean")



import os
import json
import requests
from typing import Optional
from krixik.system_builder.functions import fetch_output_endpoint


def download_file(url, save_path):
    try:
        with requests.get(url, timeout=60) as r:
            r.raise_for_status()
            with open(save_path, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
    except requests.exceptions.HTTPError as e:
        raise requests.exceptions.HTTPError(f"download_file failed with HTTPError exception: {e}")


def save_output(url: str, file_id: str, extension: str, output_save_directory: str) -> str:
    save_path = f"{output_save_directory}/{file_id}{extension}"
    download_file(url, save_path)
    return save_path


def fetch_output(self, file_id: str, local_save_directory: str) -> Optional[dict]:
    if file_id is None:
        raise ValueError("please provide a file_id")
    if not os.path.exists(local_save_directory):
        raise ValueError(f"local_save_directory {local_save_directory} does not exist")
    if not os.access(local_save_directory, os.W_OK):
        raise ValueError(f"local_save_directory {local_save_directory} is not writeable")
    if not os.access(local_save_directory, os.R_OK):
        raise ValueError(f"local_save_directory {local_save_directory} is not readable")

    if hasattr(self, "_KrixikBasePipeline__pipeline"):
        pipeline = self._KrixikBasePipeline__pipeline
    elif hasattr(self, "_KrixikSearchPipeline__pipeline"):
        pipeline = self._KrixikSearchPipeline__pipeline
    else:
        raise ValueError("pipeline not found in self")

    if hasattr(self, "_KrixikBasePipeline__version"):
        version = self._KrixikBasePipeline__version
    elif hasattr(self, "_KrixikSearchPipeline__version"):
        version = self._KrixikSearchPipeline__version
    else:
        raise ValueError("version not found in self")

    # prep payload data
    payload_data = {
        "pipeline": pipeline,
        "version": version,
        "file_id": file_id,
    }

    if hasattr(self, "_KrixikBasePipeline__api_key") and hasattr(self, "_KrixikBasePipeline__api_url"):
        api_key = self._KrixikBasePipeline__api_key
        api_url = self._KrixikBasePipeline__api_url
    elif hasattr(self, "_KrixikSearchPipeline__api_key") and hasattr(self, "_KrixikSearchPipeline__api_url"):
        api_key = self._KrixikSearchPipeline__api_key
        api_url = self._KrixikSearchPipeline__api_url
    else:
        raise ValueError("api_key and api_url not found in self")

    # prep headers
    headers = {"Content-Type": "text/plain", "krixikApiKey": api_key}

    try:
        # make request
        response = requests.post(
            (api_url or "") + fetch_output_endpoint,
            headers=headers,
            json=payload_data,
            timeout=60,
        )

        # return response
        results = json.loads(response.text)
        process_output = results["process_output"]
        if process_output is not None:
            save_paths = []
            for output in process_output:
                save_path = save_output(
                    output["url"],
                    file_id,
                    output["extension"],
                    local_save_directory,
                )
                save_paths.append(save_path)
            results["process_output_files"] = save_paths
            results["message"] += "Output saved to location(s) listed in process_output_files."

            if len(process_output) == 1:
                if process_output[0]["extension"] == ".json":
                    file_size_bytes = os.path.getsize(save_paths[0])
                    file_size_mb = file_size_bytes / (1024 * 1024)
                    if file_size_mb >= 0.5:
                        print("INFO: output json downloaded but larger than 0.5MB and will not be returned with .process output")
                        results["process_output"] = None
                    else:
                        with open(save_paths[0], "r") as file:
                            results["process_output"] = json.load(file)
                else:
                    results["process_output"] = None
            else:
                results["process_output"] = None
        status_code_dict = {"status_code": response.status_code}
        status_code_dict.update(results)
        return status_code_dict
    except requests.exceptions.HTTPError as e:
        raise requests.exceptions.HTTPError(f"FAILURE: fetch_output failed with HTTPError exception: {e}")
    except requests.exceptions.ConnectionError as e:
        raise requests.exceptions.ConnectionError(f"FAILURE: fetch_output failed with ConnectionError exception: {e}")
    except requests.exceptions.Timeout as e:
        raise requests.exceptions.Timeout(f"FAILURE: fetch_output failed with Timeout exception: {e}")
    except requests.exceptions.RequestException as e:
        raise requests.exceptions.RequestException(f"FAILURE: fetch_output failed with RequestException exception: {e}")
    except Exception as e:
        ValueError(f"FAILURE: fetch_output failed with general exception: {e}")
    finally:
        self._reset_class_variables()



import os


def local_save_directory_checker(local_save_directory: str) -> None:
    if local_save_directory is not None:
        if not isinstance(local_save_directory, str):
            raise TypeError(f"invalid local save directory: not a string - {local_save_directory}")

        if os.path.isdir(local_save_directory) is False:
            raise FileNotFoundError(f"invalid local save directory: does not exist - {local_save_directory}")

        if os.access(local_save_directory, os.W_OK) is False:
            raise PermissionError(f"invalid local save directory: write access denied - {local_save_directory}")

        if os.access(local_save_directory, os.R_OK) is False:
            raise PermissionError(f"invalid local save directory: read access denied - {local_save_directory}")



import os
import json
from typing import Optional


def is_valid(local_file_path: str) -> Optional[bool]:
    try:
        # Read the content of the JSON file and load it
        with open(local_file_path, "r") as file:
            json_object = json.load(file)

        # Check if the loaded object is a dictionary
        if isinstance(json_object, list):
            # check that json_object has at least one key
            if len(json_object) < 1:
                raise ValueError("JSON file does not represent a dictionary.")
            for v in json_object:
                if not isinstance(v, dict):
                    raise ValueError(f"object in JSON file does not represent a valid dictionary: {v}")
            return True
        else:
            raise ValueError("JSON file must consist of a list of dictionaries.")
    except json.JSONDecodeError as e:
        raise ValueError(f"Invalid JSON file - your file is not a valid JSON file: {e}")
    except FileNotFoundError:
        raise FileNotFoundError(f"The file '{local_file_path}' does not exist.")
    except Exception as e:
        raise ValueError(f"Error reading JSON file: {e}")


def is_size(
    *,
    local_file_path: str,
    snippet_min_token_count: int = 1,
    snippet_max_token_count: int = 255,
    snippet_max_key_length: int = 255,
    snippet_max_count: int = 1000000,
    minimum_file_size: float = 0.000001,
    maximum_file_size: float = 3.000001,
) -> None:
    # proper size
    def compute_size(file_path: str):
        try:
            # Get the size of the file in bytes
            file_size_bytes = os.path.getsize(file_path)

            # Convert the size to megabytes (MB)
            file_size_mb = file_size_bytes / (1024 * 1024)
            return file_size_mb
        except Exception as e:
            raise ValueError(f"text size calculation failed with exception {e}")

    # compute snippet token length and count
    def compute_snippet_size(file_path: str):
        try:
            minimum_token_count = 1
            minimum_token_count_key_value = {}
            maximum_token_count = 0
            maximum_token_count_key_value = {}
            maximum_key_length = 0
            largest_key = ""
            with open(file_path, "r") as file:
                data = json.load(file)
                snippet_count = len(data)
                for datapoint in data:
                    counter = 0
                    for key, value in datapoint.items():
                        try:
                            if len(key) > maximum_key_length:
                                maximum_key_length = len(key)
                                largest_key = key
                            if isinstance(value, str):
                                token_count = len(value.split())
                                if counter == 0:
                                    minimum_token_count = token_count
                                    maximum_token_count = token_count
                                    minimum_token_count_key_value = {key: value}
                                    maximum_token_count_key_value = {key: value}

                                if token_count > maximum_token_count:
                                    maximum_token_count = len(value.split())
                                    maximum_token_count_key_value = {key: value}

                                if token_count < minimum_token_count:
                                    minimum_token_count = token_count
                                    minimum_token_count_key_value = {key: value}

                                if len(value.strip()) == 0:
                                    minimum_token_count = 0
                                    minimum_token_count_key_value = {key: value}
                            counter += 1
                        except Exception as e:
                            raise ValueError(f"snippet token and count calculation failed on key-value pair {key}-{value} with exception {e}")

                return (
                    maximum_key_length,
                    largest_key,
                    minimum_token_count,
                    maximum_token_count,
                    snippet_count,
                    minimum_token_count_key_value,
                    maximum_token_count_key_value,
                )
        except Exception as e:
            raise ValueError(f"snippet token and count calculation failed with exception {e}")

    try:
        # check that local_file_path represents a valid json file
        is_valid(local_file_path)

        # check size of input json file
        file_size = compute_size(local_file_path)

        # check that file size in megabytes is greater than minimum_file_size and less than maximum_file_size
        if file_size < minimum_file_size:
            raise ValueError(
                f"input file size is {round(file_size,2)} megabytes: less than {minimum_file_size} megabytes (current minimum size allowable)"
            )

        if file_size > maximum_file_size:
            raise ValueError(
                f"*** Krixik Open Beta warning *** input file size is {round(file_size,2)} megabytes: greater than {maximum_file_size} megabytes (current maximum size allowable) - {local_file_path} during Krixik Open Beta"
            )

        (
            maximum_key_length,
            largest_key,
            minimum_token_count,
            maximum_token_count,
            snippet_count,
            minimum_token_count_key_value,
            maximum_token_count_key_value,
        ) = compute_snippet_size(local_file_path)

        if maximum_key_length > snippet_max_key_length:
            raise ValueError(
                f"too long key - the following key {largest_key} is too long (at present the maximum is {snippet_max_key_length} - please check this key in your input local_file_path {local_file_path}"
            )

        if snippet_count > snippet_max_count:
            raise ValueError(
                f"there are too many key-value pairs in your input local_file_path {local_file_path} - total count must be less than {snippet_max_count}"
            )
        if minimum_token_count < snippet_min_token_count:
            raise ValueError(
                f"too few tokens - the value in the following key-value pair has too few tokens (at present the minimum is {snippet_min_token_count} - please check this key-value pair in your input local_file_path {local_file_path} - {minimum_token_count_key_value}"
            )
        if maximum_token_count > snippet_max_token_count:
            raise ValueError(
                f"too many tokens - the value in the following key-value pair has too many tokens (at present the maximum is {snippet_max_token_count} - please check this key-value pair in your input local_file_path {local_file_path} - {maximum_token_count_key_value} "
            )

    except ValueError as ve:
        raise ve
    except Exception as e:
        raise ValueError(f"invalid local_file_path: {e}")



import os
from typing import Optional
# from pydub import AudioSegment
# from moviepy.editor import AudioFileClip
# from pydub.exceptions import PydubException


# def is_valid(local_file_path: str | None) -> None:
#     if local_file_path is None:
#         raise ValueError("input file is None")

#     # check that local_file_path represents audio file
#     try:
#         duration = 100
#         AudioSegment.from_file(local_file_path, format="mp3", duration=duration)
#     except PydubException as e:
#         raise ValueError(f"input file is not a valid audio file - {local_file_path} - see exception {e}")
#     except FileNotFoundError:
#         raise FileNotFoundError(f"The file '{local_file_path}' does not exist.")
#     except Exception as e:
#         raise ValueError(f"Error reading audio: {e}")


def is_valid(local_file_path: Optional[str]) -> None:
    if local_file_path is None:
        raise ValueError("input local_file_path is None")
    if not isinstance(local_file_path, str):
        raise ValueError(f"input local_file_path is not a string: '{local_file_path}'")
    if not local_file_path.endswith("mp3"):
        raise ValueError(f"input local_file_path does not end with required extension 'mp3' {local_file_path}")


def is_size(
    *,
    local_file_path: str,
    minimum_seconds: int = 1,
    maximum_seconds: int = 180,
    minimum_file_size: float = 0.00001,
    maximum_file_size: float = 3.000001,
) -> None:
    # proper size
    def compute_size(file_path: str):
        try:
            # Get the size of the file in bytes
            file_size_bytes = os.path.getsize(file_path)

            # Convert the size to megabytes (MB)
            file_size_mb = file_size_bytes / (1024 * 1024)
            return file_size_mb
        except Exception as e:
            raise ValueError(f"audio extraction failed with exception {e}")

    # def compute_duration(file_path: str) -> float:
    #     try:
    #         audio_clip = AudioFileClip(file_path)
    #         duration_in_seconds = round(audio_clip.duration, 1)
    #         audio_clip.close()
    #         return duration_in_seconds
    #     except Exception as e:
    #         raise ValueError(f"audio duration calculation failed with exception {e}")

    try:
        # check that local_file_path is valid
        is_valid(local_file_path)

        # compute file size
        file_size = compute_size(local_file_path)

        # check that file size in megabytes is greater than minimum_file_size and less than maximum_file_size
        if file_size < minimum_file_size:
            raise ValueError(
                f"input file size is {round(file_size,2)} megabytes: less than {minimum_file_size} megabytes (current minimum size allowable)"
            )

        if file_size > maximum_file_size:
            raise ValueError(
                f"***Krixik Open Beta warning*** input file size is {round(file_size,2)} megabytes: greater than {maximum_file_size} megabytes (current maximum size allowable) - {local_file_path} during Krixik Open Beta"
            )

        # # compute length of audio file in seconds
        # audio_duration = compute_duration(local_file_path)

        # # check that audio file is  greater than minimum_seconds and less than maximum_seconds
        # if audio_duration < minimum_seconds or audio_duration > maximum_seconds:
        #     raise ValueError(
        #         f"audio file is {round(audio_duration,2)} seconds - this is either less than {minimum_seconds} seconds (current minimum size allowable) or greater than {maximum_seconds} seconds (current maximum size allowable)"
        #     )
    except ValueError as ve:
        raise ve
    except Exception as e:
        raise ValueError(f"audio extraction failed with exception {e}")



from krixik.utilities.validators.system.base.local_file_path import (
    local_file_path_checker as local_file_path_checker_utility,
)
from krixik.utilities.validators.system.data.npy import valid_extensions


def local_file_path_checker(local_file_path: str) -> None:
    # start with utility function
    local_file_path_checker_utility(local_file_path)

    # check that file_name ends with an acceptable extension
    if "." + local_file_path.split(".")[-1] not in valid_extensions:
        raise ValueError(
            f"invalid local_file_path: please check that it ends with an allowed extension which include:  {valid_extensions} - your input local_file_path is: {local_file_path}"
        )



import json
import requests
from typing import Tuple
from krixik.system_builder.functions import upload_presigned_endpoint
from krixik.system_builder.functions.delete import delete_server_files


def process_local_file(self) -> Tuple[bool, requests.Response]:
    if hasattr(self, "_KrixikBasePipeline__presigned_post_url_results"):
        if self._KrixikBasePipeline__presigned_post_url_results is None:
            raise ValueError("presigned_post_url_results is None - cannot upload file for processing")
        if "url" not in self._KrixikBasePipeline__presigned_post_url_results:
            raise ValueError("key 'url' not in presigned_post_url_data")
        if "fields" not in self._KrixikBasePipeline__presigned_post_url_results:
            raise ValueError("key 'fields' not in presigned_post_url_data")
        url = self._KrixikBasePipeline__presigned_post_url_results["url"]
        fields = self._KrixikBasePipeline__presigned_post_url_results["fields"]
    elif hasattr(self, "_KrixikSearchPipeline__presigned_post_url_results"):
        if self._KrixikSearchPipeline__presigned_post_url_results is None:
            raise ValueError("presigned_post_url_results is None - cannot upload file for processing")
        if "url" not in self._KrixikSearchPipeline__presigned_post_url_results:
            raise ValueError("key 'url' not in presigned_post_url_data")
        if "fields" not in self._KrixikSearchPipeline__presigned_post_url_results:
            raise ValueError("key 'fields' not in presigned_post_url_data")
        url = self._KrixikSearchPipeline__presigned_post_url_results["url"]
        fields = self._KrixikSearchPipeline__presigned_post_url_results["fields"]
    else:
        raise ValueError("presigned_post_url_results is not in self - cannot upload file for processing")

    upload_response_success = False
    try:
        if self.local_file_path is None:
            raise ValueError("local_file_path is None - cannot upload file for processing")

        # open file as binary and attempt post to s3
        with open(self.local_file_path, "rb") as f:
            files = {"file": (self.local_file_path, f)}
            upload_response = requests.post(url, data=fields, files=files, timeout=45)
            upload_response_success = True
    except requests.exceptions.HTTPError as e:
        raise requests.exceptions.HTTPError(f"local file upload request with upload request_id {self.process_id} failed with HTTPError {e}")
    except requests.exceptions.ConnectionError as e:
        raise requests.exceptions.ConnectionError(
            f"local file upload request with upload request_id {self.process_id} failed with ConnectionError {e}"
        )
    except requests.exceptions.Timeout as e:
        raise requests.exceptions.Timeout(f"local file upload request with upload request_id {self.process_id} failed with Timeout {e}")
    except requests.exceptions.RequestException as e:
        raise requests.exceptions.RequestException(
            f"local file upload request with upload request_id {self.process_id} failed with request exception {e}"
        )
    except Exception as e:
        raise Exception(f"local file upload request with upload request_id {self.process_id} failed with exception {e}")
    finally:
        if not upload_response_success:
            delete_server_files(self)
    return upload_response_success, upload_response


def get_presigned_url(self, payload_data):
    try:
        if hasattr(self, "_KrixikBasePipeline__api_key") and hasattr(self, "_KrixikBasePipeline__api_url"):
            api_key = self._KrixikBasePipeline__api_key
            api_url = self._KrixikBasePipeline__api_url
        elif hasattr(self, "_KrixikSearchPipeline__api_key") and hasattr(self, "_KrixikSearchPipeline__api_url"):
            api_key = self._KrixikSearchPipeline__api_key
            api_url = self._KrixikSearchPipeline__api_url
        else:
            raise ValueError("api_key and api_url not found in self")

        # prep headers
        headers = {"Content-Type": "text/plain", "krixikApiKey": api_key}

        # make request
        response = requests.post(
            api_url + upload_presigned_endpoint,
            headers=headers,
            json=payload_data,
            timeout=15,
        )
        result = json.loads(response.text)

        if response.status_code == 200:
            return True, result
        if response.status_code == 400:
            return False, result
        if response.status_code == 500:
            raise requests.exceptions.HTTPError("FAILURE: request check in failed with status code 500")
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 400:
            raise ValueError(json.loads(e.response.text)["message"])
        else:
            raise requests.exceptions.HTTPError(f"request failed with HTTPError {e}")
    except requests.exceptions.ConnectionError as e:
        raise requests.exceptions.ConnectionError(f"request failed with ConnectionError {e}")
    except requests.exceptions.Timeout as e:
        raise requests.exceptions.Timeout(f"request failed with Timeout {e}")
    except requests.exceptions.RequestException as e:
        raise requests.exceptions.RequestException(f"request failed with request exception {e}")
    except Exception as e:
        raise Exception(f"request failed with exception {e}")



def symbolic_path_splitter(symbolic_file_path: str) -> tuple:
    if not isinstance(symbolic_file_path, str):
        raise ValueError("symbolic_file_path must be a string")

    split = symbolic_file_path.split("/")
    file_name = split[-1]

    symbolic_directory_path = "/".join(split[:-1])
    if len(symbolic_directory_path) == 0:
        symbolic_directory_path = "/"
    return symbolic_directory_path, file_name



list_endpoint = "/krixik-core-list"
update_endpoint = "/krixik-core-update"
delete_endpoint = "/krixik-core-delete"
show_tree_endpoint = "/krixik-core-show-tree"
upload_status_endpoint = "/krixik-core-process-status"
upload_presigned_endpoint = "/krixik-core-process"
fetch_output_endpoint = "/krixik-core-fetch-output"
cli_checkin_endpoint = "/cli-checkin"
keyword_search_endpoint = "/keyword-search"
semantic_search_endpoint = "/semantic-search"
semantic_search_endpoint = "/semantic-search"
cap_endpoint = "/krixik-core-cap-check"



import copy
import re
from krixik.utilities.utilities import vprint
from krixik.utilities.validators.system import (
    MAX_FILE_NAME_LENGTH,
    MAX_FILE_NAME_COUNT,
)

# define pattern for valid file names
file_names_pattern = r"^[a-zA-Z0-9 _,.-]+(\.[a-zA-Z0-9]+)?$"


def is_valid_file_name(file_name: str) -> bool:
    return bool(re.match(file_names_pattern, file_name))


# individual file_name checker
def individual_file_name_checker(file_name: str, valid_extensions: list = [], phase: str = "other") -> None:
    if file_name is not None:
        # check that file_name is a string
        if not isinstance(file_name, str):
            raise TypeError("file_name must be input as a string")

        # check that file_name length is greater than 0
        if len(file_name) == 0:
            raise ValueError("invalid file_name: length is 0")

        # check that file_name length is less than 64
        if len(file_name) > MAX_FILE_NAME_LENGTH:
            raise ValueError(
                f"invalid file_name: length is greater than 64 (current maximum length allowable) - {file_name} - length is {len(file_name)}"
            )

        # check for single asterisk
        if file_name == "*":
            raise ValueError("invalid file_name: cannot be '*'")

        # check that file_name is a valid file name
        if phase == "process":
            res = is_valid_file_name(file_name)
            if res is False:
                raise ValueError(
                    f"invalid file_name: please check that it contains only alphanumeric, alphanumeric, '_', or '-', characters - {file_name}"
                )

            # check if * is in file_name
            if "*" in file_name:
                raise ValueError(f"invalid file_name for process: please check that it does not contain any '*' characters - {file_name}")

            # check that file_name ends with an valid_extensions
            if "." + file_name.split(".")[-1] not in valid_extensions:
                raise ValueError(
                    f"invalid file_name: please check that it ends with an allowed extension - {file_name} - allowed extensions: {valid_extensions}"
                )

        if phase != "process":
            # check if input file_name begins or ends with '*'
            test_file_name = copy.deepcopy(file_name)

            # check for **
            if "**" in test_file_name:
                raise ValueError(f"invalid file_name: please check that it does not contain any '**' characters - {file_name}")

            # check for ***
            if "***" in test_file_name:
                raise ValueError(f"invalid file_name: please check that it does not contain any '***' characters - {file_name}")

            # check if input file_name begins or ends with '*'
            test_file_name = copy.deepcopy(file_name)

            if file_name[0] == "*":
                test_file_name = test_file_name[1:]

            if file_name[-1] == "*":
                test_file_name = test_file_name[:-1]
            else:
                # extension must fall into valid_extensions
                if "." + file_name.split(".")[-1] not in valid_extensions:
                    raise ValueError(
                        f"invalid file_name: please check that it ends with an allowed extension - {file_name} - allowed extensions: {valid_extensions}"
                    )

            # split on possible extension
            non_extension = test_file_name.split(".")[0]

            if len(non_extension) > 0:
                # test for validity of file_name
                res = is_valid_file_name(non_extension)
                if res is False:
                    raise ValueError(
                        f"invalid file_name: please check that the core (excluding an optional single '*' at the beginning and/or end) of this file_name contains only basic alphanumeric, underscore, and/or hyphen characters - your input file_name: {file_name}"
                    )


# check file names
def file_names_checker(
    file_names: list,
    phase: str = "other",
    valid_extensions: list = [],
    verbose: bool = True,
) -> None:
    if file_names is not None:
        # check that file_names is a list
        if not isinstance(file_names, list):
            raise TypeError("file_names is not a list")

        # check that file_names is not empty
        if len(file_names) == 0:
            raise ValueError("file_names is empty")

        # check that file_names is not too long
        if len(file_names) > MAX_FILE_NAME_COUNT:
            raise ValueError(f"file_names contains more than {MAX_FILE_NAME_COUNT} file_names")

        # check that file_names is not empty
        if file_names is not None:
            # check that all file_names are valid file names
            for file_name in file_names:
                # check validity of file_name
                individual_file_name_checker(file_name, valid_extensions, phase=phase)

            # check for duplicates
            if len(file_names) != len(set(file_names)):
                # find duplicates
                duplicates = []
                for file_name in file_names:
                    if file_names.count(file_name) > 1:
                        duplicates.append(file_name)

                # print error message
                vprint(
                    f"WARNING: file_names contains duplicate entries - {duplicates}",
                    verbose=verbose,
                )
                vprint(
                    "INFO: You can remove duplicates by using Python's list(set()) function on your list of file_names",
                    verbose=verbose,
                )
        else:
            raise ValueError("file_names is None")



import copy
from typing import Optional, Tuple
from datetime import datetime, timezone
from krixik.utilities.utilities import vprint

# valid timestamp formats
timestamp_valid_formats = ["%Y-%m-%d", "%Y-%m-%d %H:%M:%S"]


# check that timestamp represents a valid date
def is_valid_timestamp(timestamp: str) -> None:
    for format_str in timestamp_valid_formats:
        try:
            datetime.strptime(timestamp, format_str)
        except ValueError:
            raise ValueError(f"{timestamp} must be a valid date but was given as {timestamp}")


def numerical_timestamp_checker(timestamp_int: int, timestamp_name: str) -> None:
    if not isinstance(timestamp_int, int):
        raise TypeError(f"{timestamp_name} must be an integer but was given as {timestamp_int}")
    if timestamp_int < 0:
        raise ValueError(f"{timestamp_name} must be a positive integer but was given as {timestamp_int}")


def string_timestamp_to_int(timestamp_str: str, timestamp_name: str) -> Optional[int]:
    dt_object = None
    dt_object_utc = None
    dt_check = 0
    for format in timestamp_valid_formats:
        try:
            # Attempt to parse the timestamp string into a datetime object
            dt_object = datetime.strptime(timestamp_str, format)

            # Convert the datetime object to a UTC datetime object
            dt_object_utc = dt_object.replace(tzinfo=timezone.utc)

            dt_check += 1
            break
        except ValueError:
            pass
            # raise ValueError(f'{timestamp_name} must be in format YYYY-MM-DD or YYYY-MM-DD HH:MM:SS but was given as {timestamp_str}')

    if dt_check == 0:
        raise ValueError(f"{timestamp_name} must be in format -- YYYY-MM-DD -- or -- YYYY-MM-DD HH:MM:SS -- but was given as {timestamp_str}")

    # Convert the datetime object to a Unix timestamp (seconds since epoch)
    if dt_object_utc is not None:
        timestamp_int_utc = int(dt_object_utc.timestamp())
        return timestamp_int_utc


def is_valid_format(timestamp_str: str, timestamp_name: str) -> None:
    # check if timestamp is in correct format
    dt_check = 0
    for format in timestamp_valid_formats:
        try:
            # Attempt to parse the timestamp string
            datetime.strptime(timestamp_str, format)
            dt_check += 1
        except ValueError:
            pass
            # raise ValueError(f'{timestamp_name} must be in format YYYY-MM-DD or YYYY-MM-DD HH:MM:SS but was given as {timestamp_str}')
    if dt_check == 0:
        raise ValueError(f"{timestamp_name} must be in format YYYY-MM-DD or YYYY-MM-DD HH:MM:SS but was given as {timestamp_str}")


def is_stump(timestamp_str: str, start_end: Optional[str] = None, verbose: bool = False) -> str:
    try:  # if timestamp_str is format '%Y-%m-%d' then add '23:59:59' to the end of the timestamp
        datetime.strptime(timestamp_str, "%Y-%m-%d")
        og_timestamp_str = copy.deepcopy(timestamp_str)
        if start_end == "start":
            timestamp_str = timestamp_str + " 00:00:00"
        if start_end == "end":
            timestamp_str = timestamp_str + " 23:59:59"
        vprint(
            f"INFO: Interpreting {og_timestamp_str} as {timestamp_str} to ensure temporally inclusive query",
            verbose=verbose,
        )
        return timestamp_str
    except ValueError:
        return timestamp_str


def is_after_bot(timestamp: str, timestamp_name: str) -> None:
    # check that timestamp occurs after 2023-01-01
    if timestamp < "2023-01-01":
        raise ValueError(f"{timestamp_name} must be after 2023-01-01 but was given as {timestamp}")

    # check if timestamp occurs after the year 9999
    if timestamp > "9999-12-31":
        raise ValueError(f"{timestamp_name} must be before 9999-12-31 but was given as {timestamp}")


def timestamp_checker(timestamp: str, timestamp_name: str, start_end: str, verbose: bool = False) -> Optional[int]:
    # check if timestamp is a string
    if not isinstance(timestamp, str):
        raise TypeError(f"{timestamp_name} must be a string but was given as {timestamp}")

    # check if timestamp is in correct format
    is_valid_format(timestamp, timestamp_name)

    # check if stump
    timestamp = is_stump(timestamp, start_end=start_end, verbose=verbose)

    # check that timestamp occurs after 2023-01-01
    is_after_bot(timestamp, timestamp_name)

    # convert timestamp to int
    timestamp_int = string_timestamp_to_int(timestamp, timestamp_name)

    # double check timestamp_int is a positive integer
    if timestamp_int is not None:
        numerical_timestamp_checker(timestamp_int, timestamp_name)
        return timestamp_int


def timestamp_bookends_checker(
    created_at_start: Optional[str] = None,
    created_at_end: Optional[str] = None,
    last_updated_start: Optional[str] = None,
    last_updated_end: Optional[str] = None,
    verbose: bool = False,
) -> None:
    # make copy for reference messaging
    og_created_at_start = copy.deepcopy(created_at_start)
    og_created_at_end = copy.deepcopy(created_at_end)
    og_last_updated_start = copy.deepcopy(last_updated_start)
    og_last_updated_end = copy.deepcopy(last_updated_end)

    # check each bookend
    created_at_start_int = None
    created_at_end_int = None
    last_updated_start_int = None
    last_updated_end_int = None
    if created_at_start is not None:
        created_at_start_int = timestamp_checker(created_at_start, "created_at_start", start_end="start", verbose=verbose)

    if created_at_end is not None:
        created_at_end_int = timestamp_checker(created_at_end, "created_at_end", start_end="end", verbose=verbose)

    if last_updated_start is not None:
        last_updated_start_int = timestamp_checker(last_updated_start, "last_updated_start", start_end="start", verbose=verbose)

    if last_updated_end is not None:
        last_updated_end_int = timestamp_checker(last_updated_end, "last_updated_end", start_end="end", verbose=verbose)

    # check comparisons between bookends
    if created_at_start_int is not None and created_at_end_int is not None:
        if created_at_start_int > created_at_end_int:
            raise ValueError(f"created_at_start {og_created_at_start} must occur before created_at_end {og_created_at_end}.")

    if last_updated_start_int is not None and last_updated_end_int is not None:
        if last_updated_start_int > last_updated_end_int:
            raise ValueError(f"last_updated_start {og_last_updated_start} must occur before last_updated_end {og_last_updated_end}.")


def convert_timestamps(
    created_at_start: Optional[str],
    created_at_end: Optional[str],
    last_updated_start: Optional[str],
    last_updated_end: Optional[str],
) -> Tuple[Optional[int], Optional[int], Optional[int], Optional[int]]:
    created_at_start_int = None
    created_at_end_int = None
    last_updated_start_int = None
    last_updated_end_int = None
    if created_at_start is not None:
        created_at_start = is_stump(created_at_start, "start", verbose=False)
        created_at_start_int = string_timestamp_to_int(created_at_start, "created_at_start")
    if last_updated_start is not None:
        last_updated_start = is_stump(last_updated_start, "start", verbose=False)
        last_updated_start_int = string_timestamp_to_int(last_updated_start, "last_updated_start")
    if created_at_end is not None:
        created_at_end = is_stump(created_at_end, "end", verbose=False)
        created_at_end_int = string_timestamp_to_int(created_at_end, "created_at_end")
    if last_updated_end is not None:
        last_updated_end = is_stump(last_updated_end, "end", verbose=False)
        last_updated_end_int = string_timestamp_to_int(last_updated_end, "last_updated_end")
    return (
        created_at_start_int,
        created_at_end_int,
        last_updated_start_int,
        last_updated_end_int,
    )



import json
from typing import Dict, Any
import requests
import time
from typing import Optional, Union
from krixik.system_builder.functions import upload_status_endpoint
from krixik.utilities.utilities import vprint


def process_status_reporter(
    self,
    prev_process_status: Optional[Union[str, dict]],
    process_status: Optional[Union[str, dict]],
    verbose: bool,
) -> bool:
    if process_status is None or (isinstance(process_status, str)):
        raise ValueError(f"file process failed - the request_id of this failed process is {self.process_id}")

    # loop through og_process_status and check if any processing_complete is True
    module_count = 1
    for key, value in process_status.items():
        if process_status[key]:
            if prev_process_status is None or not prev_process_status[key]:
                vprint(
                    f"SUCCESS: module {module_count} (of {len(process_status)}) - {key.split('.')[0]} processing complete.",
                    verbose=verbose,
                )
        module_count += 1
    # if all values in process_status are True, then process is complete
    values = list(process_status.values())
    process_complete = True
    if not all(values):
        process_complete = False

    return process_complete


def check_process_status(self, *, process_id: str) -> tuple:
    max_count = 9
    start_count = 0
    backoff_schedule = [1, 2, 4, 8, 16, 32, 64, 128, 256]
    while start_count < max_count:
        try:
            if process_id is None:
                raise ValueError("request_id cannot be none when checking process_status")

            if hasattr(self, "_KrixikBasePipeline__pipeline"):
                pipeline = self._KrixikBasePipeline__pipeline
            elif hasattr(self, "_KrixikSearchPipeline__pipeline"):
                pipeline = self._KrixikSearchPipeline__pipeline
            else:
                raise ValueError("pipeline not found in self")

            if hasattr(self, "_KrixikBasePipeline__version"):
                version = self._KrixikBasePipeline__version
            elif hasattr(self, "_KrixikSearchPipeline__version"):
                version = self._KrixikSearchPipeline__version
            else:
                raise ValueError("version not found in self")

            # prep payload data
            payload_data: Dict[str, str] = {
                "pipeline": pipeline,
                "version": version,
                "process_id": process_id,
            }

            if hasattr(self, "_KrixikBasePipeline__api_key") and hasattr(self, "_KrixikBasePipeline__api_url"):
                api_key = self._KrixikBasePipeline__api_key
                api_url = self._KrixikBasePipeline__api_url
            elif hasattr(self, "_KrixikSearchPipeline__api_key") and hasattr(self, "_KrixikSearchPipeline__api_url"):
                api_key = self._KrixikSearchPipeline__api_key
                api_url = self._KrixikSearchPipeline__api_url
            else:
                raise ValueError("api_key and api_url not found in self")

            # prep headers - type hint correctly
            headers: Dict[str, str] = {
                "Content-Type": "text/plain",
                "krixikApiKey": api_key or "",
            }

            # make request
            response: requests.Response = requests.post(
                (api_url or "") + upload_status_endpoint,
                headers=headers,
                json=payload_data,
                timeout=300,
            )

            # unpack response
            response_text = json.loads(response.text)
            self.request_id = response_text["request_id"]

            # Check the response
            if response.status_code == 200:
                file_id = response_text["file_id"]
                process_status: Dict[str, Any] = json.loads(response.text)["process_status"]

                failure_status: Any = response_text["failure_status"]
                message = response_text["message"]
                return file_id, process_status, failure_status, message
            else:
                message = response_text["message"]
                raise ValueError(
                    f"request failed with status code: {response.status_code} - the request_id of this failed status check is {self.request_id} - this request failed with the following message from the server: {message}"
                )
        except ValueError as err:
            raise err
        except requests.exceptions.ConnectionError as e:
            if start_count < max_count:
                time.sleep(backoff_schedule[start_count])
                start_count += 1
                continue
            raise requests.exceptions.ConnectionError(
                f"request failed with ConnectionError {e} - the request_id of this failed status check is {self.request_id}"
            )
        except requests.exceptions.Timeout as e:
            if start_count < max_count:
                time.sleep(backoff_schedule[start_count])
                start_count += 1
                continue
            raise requests.exceptions.Timeout(f"request failed with Timeout {e} - the request_id of this failed status check is {self.request_id}")
        except requests.exceptions.RequestException as e:
            raise requests.exceptions.RequestException(
                f"request failed with request exception {e} - the request_id of this failed status check is {self.request_id}"
            )



import re
from krixik.utilities.validators.system import MAX_FILE_ID_COUNT
from krixik.utilities.utilities import vprint

# define pattern for valid file id
file_id_pattern = r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$"


def is_valid_uuid_string(input_string: str) -> bool:
    return bool(re.match(file_id_pattern, input_string))


def individual_file_id_checker(file_id: str) -> None:
    if file_id is not None:
        # check that file_id is a string
        if not isinstance(file_id, str):
            if file_id is None:
                raise ValueError("file_id must be input as an argument")
            else:
                raise TypeError("file_id must be input as a string")

        # check that file_id matches pattern
        res = is_valid_uuid_string(file_id)
        if res is False:
            raise TypeError(f"invalid file_id: a valid file_id is a string of length 36 with the following pattern: {file_id_pattern}")

        # check that file_id length is 36
        if len(file_id) != 36:
            raise TypeError(f"invalid file_id: length of this file_id is not 36 (it must be exactly 36 characters) - {file_id}")


def file_ids_checker(file_ids: list, verbose: bool = True) -> None:
    if file_ids is not None:
        # check that file_ids is a list of strings
        if not isinstance(file_ids, list):
            raise TypeError("file_ids is not a list")

        # check that file_ids is not empty
        if file_ids is not None:
            # check if length 0 list
            if len(file_ids) == 0:
                raise ValueError("file_ids is an empty list")

            # check if length greater than MAX_FILE_ID_COUNT
            if len(file_ids) > MAX_FILE_ID_COUNT:
                raise ValueError(f"file_ids contains more than {MAX_FILE_ID_COUNT} file_ids")

            # check each file_id for validity
            for file_id in file_ids:
                # check validity of file_id
                individual_file_id_checker(file_id)

            # check for duplicates
            if len(file_ids) != len(set(file_ids)):
                # find duplicates
                duplicates = []
                for file_id in file_ids:
                    if file_ids.count(file_id) > 1:
                        duplicates.append(file_id)

                # print error message
                vprint(
                    f"WARNING: file_ids contains duplicate entries - {duplicates}",
                    verbose=verbose,
                )
                vprint(
                    "INFO: You can remove duplicates by using Python's list(set()) function on your list of file_names",
                    verbose=verbose,
                )

        else:
            raise ValueError("file_ids is None")



from krixik.pipeline_builder.module import Module
from krixik.pipeline_builder import MAX_MODULES


def chain_check(module_chain: list) -> None:
    if not isinstance(module_chain, list):
        raise TypeError(f"FAILURE: module_chain - {module_chain} - is not a list")
    if len(module_chain) == 0:
        raise ValueError(f"FAIULRE: module_chain - {module_chain} is empty")
    for item in module_chain:
        if not isinstance(item, Module):
            raise TypeError(f"FAILURE: item in module_chain - {item} - is not a proper Module object")

    if len(module_chain) > MAX_MODULES:
        raise ValueError(f"pipelines cannot currently have more than {MAX_MODULES} modules")



from krixik.__base__ import library_base_dir
from pathlib import Path
from krixik.modules.utilities.read_config import (
    get_module_available_defaults,
)

module_path_name = Path(__file__).parent.name
module_config = get_module_available_defaults(library_base_dir + f"/modules/{module_path_name}/module.yml")



from functools import wraps


allowed_args_dict = {
    "init": {"api_key", "api_url"},
    "delete": {"file_id"},
    "list": {
        "file_ids",
        "file_names",
        "symbolic_directory_paths",
        "symbolic_file_paths",
        "file_tags",
        "sort_order",
        "max_files",
        "created_at_start",
        "created_at_end",
        "last_updated_start",
        "last_updated_end",
        "verbose",
    },
    "fetch_output": {"file_id", "local_save_directory"},
    "update": {
        "file_id",
        "file_name",
        "symbolic_directory_path",
        "symbolic_file_path",
        "file_tags",
        "file_description",
        "expire_time",
        "verbose",
    },
    "process": {
        "file_id",
        "file_name",
        "symbolic_directory_path",
        "symbolic_file_path",
        "local_file_path",
        "file_tags",
        "file_description",
        "modules",
        "expire_time",
        "verbose",
        "wait_for_process",
        "local_save_directory",
        "og_local_file_path",
        "download_output",
    },
    "show_tree": {"symbolic_directory_path", "max_files", "verbose"},
    "process_status": {"request_id"},
    "keyword_search": {
        "query",
        "file_ids",
        "file_names",
        "symbolic_directory_paths",
        "symbolic_file_paths",
        "file_tags",
        "sort_order",
        "max_files",
        "created_at_start",
        "created_at_end",
        "last_updated_start",
        "last_updated_end",
        "verbose",
    },
    "semantic_search": {
        "query",
        "file_ids",
        "file_names",
        "symbolic_directory_paths",
        "symbolic_file_paths",
        "file_tags",
        "sort_order",
        "max_files",
        "created_at_start",
        "created_at_end",
        "last_updated_start",
        "last_updated_end",
        "verbose",
        "k",
    },
}


def kwargs_checker(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        func_name = func.__name__
        if func_name == "converter_wrapper":
            func_name = "process"
        allowed_args = allowed_args_dict.get(func_name)
        if allowed_args is None:
            raise ValueError(f"function not supported: '{func_name}'")

        unexpected_args = set(kwargs.keys()) - allowed_args
        if unexpected_args:
            raise TypeError(f"unexpected keyword argument(s) for '{func_name}': {', '.join(unexpected_args)}")
        return func(*args, **kwargs)

    return wrapper



from typing import Optional
import yaml
from krixik.__base__ import library_base_dir

config_path = f"{library_base_dir}/utilities/converters/utilities/config.yml"


def convert_extension(extension: str) -> Optional[dict]:
    with open(config_path, "r") as file:
        data_config = yaml.safe_load(file)["convertable_extensions"]
    simple_data_config = {k["name"]: k["convert"] for k in data_config}
    if extension not in list(simple_data_config.keys()):
        return None
    return simple_data_config[extension]



from typing import Optional
from krixik.__version__ import __version__
import tempfile
import types
from krixik.utilities.utilities import classproperty
from krixik.modules import available_modules, get_module_details
from krixik.system_builder.functions.checkin import checkin
from krixik.system_builder.functions.cap_check import cap_check
from krixik.pipeline_builder.pipeline import BuildPipeline
from krixik.system_builder.base import KrixikBasePipeline
from krixik.system_builder.functions.semantic_search import semantic_search
from krixik.system_builder.functions.keyword_search import keyword_search
from krixik.pipeline_builder.utilities.config_checker import config_check
from krixik.pipeline_builder.module import Module
from krixik.pipeline_builder.pipeline import BuildPipeline


class krixik:
    """Main class for krixik pipeline selection and initialization"""

    __api_key = None
    __api_url = None
    __api_check_val = None
    __local_conversion_directory = tempfile.gettempdir()

    @classmethod
    def init(cls, *, api_key: Optional[str], api_url: Optional[str]) -> None:
        """krixik pipeline initialization method - initializes unique user session using api_key and api_url

        Parameters
        ----------
        api_key : str
            user api key
        api_url : str
            user api url
        """
        cls.__api_key = api_key
        cls.__api_url = api_url
        cls.__api_check_val = checkin(cls.__api_key, cls.__api_url)
        cls.__version = __version__

    @classmethod
    def check_init_data(cls) -> dict:
        """convenience check for init of api_key and api_url

        Returns
        -------
        dict
            dictionary with api_key, api_url and api_check_val
        """
        return {
            "api_key": cls.__api_key,
            "api_url": cls.__api_url,
            "api_check_val": cls.__api_check_val,
        }

    @classmethod
    def create_pipeline(cls, *, name: str, module_chain: list):
        if not isinstance(name, str):
            raise TypeError("pipeline name must be a string")
        if not (len(name) > 0 and len(name) < 128):
            raise ValueError("pipeline name must have length greater than 0 and less than 128")
        if not isinstance(module_chain, list):
            raise TypeError("module_chain must be a list of strings")
        for item in module_chain:
            if not isinstance(item, str):
                raise TypeError(f"module_chain must be a list of strings - the following item in it is not a string - {item}")
            if item not in available_modules:
                raise ValueError(f"module_chain item - {item} - is not a currently one of the currently available modules -{available_modules}")
        module_chain_ = [Module(m_name) for m_name in module_chain]
        custom = BuildPipeline(name=name, module_chain=module_chain_)
        return cls.load_pipeline(pipeline=custom)

    @classmethod
    def load_pipeline(cls, *, config_path: Optional[str] = None, pipeline: Optional[BuildPipeline] = None) -> object:
        """load pipeline: from object or configuration file

        Parameters
        ----------
        config_path : str
            path to pipeline configuration file
        pipeline: BuildPipeline
            pipeline object
        """
        # only one of config_path or pipeline can be passed
        if config_path is None and pipeline is None:
            raise ValueError("config_path or pipeline must be passed")

        if config_path is not None and pipeline is not None:
            raise ValueError("only one of config_path or pipeline can be passed, not both")

        if config_path is not None:
            config_check(config_path)
            custom_pipeline = BuildPipeline(config_path=config_path)
        else:
            if not isinstance(pipeline, BuildPipeline):
                raise TypeError(f"input pipeline not proper BuildPipeline object")
            custom_pipeline = pipeline

        # pass init
        init_data = cls.check_init_data()

        # instantiate krixik server pipeline object
        pipeline_object = KrixikBasePipeline(
            pipeline=custom_pipeline.name,
            module_chain=custom_pipeline.module_chain,
            output_process_keys=custom_pipeline.module_chain_output_process_keys,
            api_key=init_data["api_key"],
            api_url=init_data["api_url"],
            api_check_val=init_data["api_check_val"],
        )

        pipeline_object.save_pipeline = custom_pipeline.save
        pipeline_object.test_input = custom_pipeline.test_input
        pipeline_object.config = custom_pipeline.config

        if custom_pipeline.module_chain[-1] == "vector-db":
            if len(custom_pipeline.module_chain) > 1:
                if custom_pipeline.module_chain[-2] == "text-embedder":
                    pipeline_object.semantic_search = types.MethodType(semantic_search, pipeline_object)

        if custom_pipeline.module_chain[-1] == "keyword-db":
            pipeline_object.keyword_search = types.MethodType(keyword_search, pipeline_object)

        return pipeline_object

    @classmethod
    def view_module_config(cls, *, module_name: str) -> dict:
        """convenience method for examinng module details

        Parameters
        ----------
        module_name : str
            module name

        Returns
        -------
        dict
            dictionary with module details
        """
        return get_module_details(module_name)

    @classmethod
    def check_cap(cls) -> Optional[dict]:
        """_summary_

        Returns
        -------
        Optional[dict]
            _description_
        """

        # make request
        return cap_check(cls.__api_key, cls.__api_url, cls.__version)

    @classmethod
    def view_module_click_data(cls, *, module_name: str) -> dict:
        return Module(module_name).click_data

    @classproperty
    def local_conversion_directory(cls):
        return cls.__local_conversion_directory

    @classproperty
    def available_modules(cls):
        return available_modules

    @classproperty
    def version(cls):
        return cls.__version

    def __getattr__(self, attr):
        raise AttributeError(f"'{self.__class__.__name__}' object instance has no attribute '{attr}'")



import json
import requests
from krixik.system_builder.functions import delete_endpoint


def delete_server_files(self) -> dict:
    try:
        if self.file_id is None:
            raise ValueError("file_id is None - please provide a file_id")

        if hasattr(self, "_KrixikBasePipeline__pipeline"):
            pipeline = self._KrixikBasePipeline__pipeline
        elif hasattr(self, "_KrixikSearchPipeline__pipeline"):
            pipeline = self._KrixikSearchPipeline__pipeline
        else:
            raise ValueError("pipeline not found in self")

        if hasattr(self, "_KrixikBasePipeline__version"):
            version = self._KrixikBasePipeline__version
        elif hasattr(self, "_KrixikSearchPipeline__version"):
            version = self._KrixikSearchPipeline__version
        else:
            raise ValueError("version not found in self")

        payload_data = {
            "pipeline": pipeline,
            "version": version,
            "file_id": self.file_id,
        }

        if hasattr(self, "_KrixikBasePipeline__api_key") and hasattr(self, "_KrixikBasePipeline__api_url"):
            api_key = self._KrixikBasePipeline__api_key
            api_url = self._KrixikBasePipeline__api_url
        elif hasattr(self, "_KrixikSearchPipeline__api_key") and hasattr(self, "_KrixikSearchPipeline__api_url"):
            api_key = self._KrixikSearchPipeline__api_key
            api_url = self._KrixikSearchPipeline__api_url
        else:
            raise ValueError("api_key and api_url not found in self")

        headers = {"Content-Type": "text/plain", "krixikApiKey": api_key}

        response = requests.post(
            (api_url or "") + delete_endpoint,
            headers=headers,
            json=payload_data,
            timeout=60,
        )

        response_text = json.loads(response.text)
        self.request_id = response_text["request_id"]

        results = json.loads(response.text)
        status_code_dict = {"status_code": response.status_code}
        status_code_dict.update(results)
        return status_code_dict
    except requests.exceptions.HTTPError as e:
        raise requests.exceptions.HTTPError(f"delete failed with HTTPError {e}")
    except requests.exceptions.ConnectionError as e:
        raise requests.exceptions.ConnectionError(f"delete failed with ConnectionError {e}")
    except requests.exceptions.Timeout as e:
        raise requests.exceptions.Timeout(f"delete failed with Timeout {e}")
    except requests.exceptions.RequestException as e:
        raise requests.exceptions.RequestException(f"delete failed with request exception {e}")
    except Exception as e:
        raise ValueError(f"delete failed with request exception {e}")
    finally:
        self._reset_class_variables()



import yaml
import os


def config_check(filename):
    if not isinstance(filename, str):
        raise TypeError(f" input config_path must be a string: {filename}")

    if not os.path.exists(filename):
        raise FileExistsError(f"FAILURE: config file does not exist - {filename}")
    if not os.path.isfile(filename):
        raise FileExistsError(f"FAILURE: config file does not exist - {filename}")

    file_size_bytes = os.path.getsize(filename)
    file_size_mb = file_size_bytes / (1024 * 1024)
    max_config_size = 1
    if file_size_mb >= max_config_size:
        raise ValueError(
            f"FAILURE: the size of your input configuration file is greater than {max_config_size}MB - the maximum size allowed - please check this file to ensure it only contains your pipeline configuration: {filename}"
        )

    _, extension = os.path.splitext(filename)
    if extension.lower() == ".yaml" or extension.lower() == ".yml":
        try:
            with open(filename, "r") as file:
                yaml.safe_load(file)
        except yaml.YAMLError:
            raise yaml.YAMLError(f"config is not a valid yaml - {filename}")
    else:
        raise ValueError(f"config - {filename} - does not end with valid extension .yml or .yaml")

    with open(filename, "r") as file:
        pipeline_config = yaml.safe_load(file)

    if "pipeline" not in list(pipeline_config.keys()):
        raise ValueError(f"did not find required 'pipeline' declaration in your input config {filename}")

    pipeline = pipeline_config["pipeline"]

    if "name" not in list(pipeline.keys()):
        raise ValueError(f"did not find required 'name' declaration in your config pipeline declaration {filename}")

    if "modules" not in list(pipeline.keys()):
        raise ValueError(f"did not find required 'modules' declaration in your config pipeline declaration {filename}")



from typing import Optional, Union


def clean_strings(input_data: Optional[Union[str, list]]):
    if input_data is not None:
        if isinstance(input_data, str):
            c_string = input_data.replace("\n", " ").replace("\r", "").strip()
            return c_string
        elif isinstance(input_data, list):
            c_strings = [v.replace("\n", " ").replace("\r", "").strip() for v in input_data]
            return c_strings



import os
from typing import Optional
from pptx import Presentation
from krixik.utilities.validators.system.base.utilities.decorators import (
    file_converters_input_check,
)
from krixik.utilities.validators.data.pptx import is_size as is_size_pptx
from krixik.utilities.validators.data.text import is_size as is_size_text
from krixik.utilities.utilities import vprint


def delete_file(*, file_path: str, exception: str):
    if file_path is not None:
        if os.path.exists(file_path):
            os.remove(file_path)
        raise ValueError(
            f"there seems to be a problem with your pptx - file conversion failed with exception {str(exception)} - please check your pptx file of the converter documentation - python-pptx - for further information: https://python-pptx.readthedocs.io/en/stable/"
        )


@file_converters_input_check
def extract(
    *,
    local_file_path: str,
    local_save_directory: Optional[str] = None,
    verbose: bool = True,
) -> Optional[str]:
    if local_file_path is None:
        raise NameError("local_file_path not defined")
    if os.path.exists(local_file_path) is False:
        raise FileNotFoundError(f"the file '{local_file_path}' does not exist.")
    if local_save_directory is None:
        raise NameError("local_save_directory is not defined")
    if os.path.exists(local_save_directory) is False:
        raise FileNotFoundError(f"the directory '{local_save_directory}' does not exist.")
    if not isinstance(verbose, bool):
        raise ValueError("verbose is not a boolean")

    convert_save_path = ""
    try:
        file_name = "krixik_converted_version_" + os.path.splitext(os.path.basename(local_file_path))[0]
        extension = ".txt"
        convert_save_path = os.path.join(local_save_directory, file_name + extension)

        # remove file if it already exists
        if os.path.exists(convert_save_path):
            os.remove(convert_save_path)

        # creating a pptx reader object
        prs = Presentation(local_file_path)

        # loop through slides, shapes, paragraphs, and runs to get all text
        num = 0
        for slide in prs.slides:
            for shape in slide.shapes:
                if not shape.has_text_frame:
                    continue
                for paragraph in shape.text_frame.paragraphs:
                    for run in paragraph.runs:
                        line = run.text
                        if num == 0:
                            with open(convert_save_path, "a", encoding="utf-8") as new_file:
                                new_file.write(line)
                        if num > 0:
                            if os.path.exists(convert_save_path):
                                with open(convert_save_path, "a", encoding="utf-8") as new_file:
                                    new_file.write(line)
                            else:
                                raise FileNotFoundError(f"file {convert_save_path} does not exist")

                        num += 1
        # report success
        vprint(
            f"SUCCESS: File conversion complete python-pptx, result saved to: {convert_save_path}",
            verbose=verbose,
        )

        # report size check
        vprint(
            "INFO: Checking that file size falls within acceptable parameters...",
            verbose=verbose,
        )
        is_size_text(local_file_path=convert_save_path)
        vprint("INFO:...success!", verbose=verbose)

        return convert_save_path
    except ValueError as ve:
        vprint(
            "INFO:...failure!  The converted (text) file does not fall within acceptable size parameters.",
            verbose=verbose,
        )
        delete_file(file_path=convert_save_path, exception=str(ve))
    except Exception as e:
        delete_file(file_path=convert_save_path, exception=str(e))


@file_converters_input_check
def convert(
    *,
    local_file_path: Optional[str],
    local_save_directory: Optional[str] = None,
    verbose: bool = True,
) -> str:
    """pptx converter - converts pptx file to txt using the pptx library

    Parameters
    ----------
    local_file_path : str | None
        path to local file to convert
    local_save_directory : str | None
        local directory to save converted file
    verbose : bool, optional
        by default True

    Returns
    -------
    str | None
        if conversion successful returns path to converted file
    """

    if local_file_path is None:
        raise NameError("local_file_path not defined")
    if os.path.exists(local_file_path) is False:
        raise FileNotFoundError(f"the file '{local_file_path}' does not exist.")

    if "." + local_file_path.split(".")[-1] == ".pptx":
        if local_save_directory is None:
            raise NameError("local_save_directory is not defined")

        vprint("INFO: Converting pptx to text...", verbose=verbose)

        # check size and validity of input pptx file
        is_size_pptx(local_file_path=local_file_path)

        # convert pptx to text
        new_local_file_path = extract(
            local_file_path=local_file_path,
            local_save_directory=local_save_directory,
            verbose=verbose,
        )

        return new_local_file_path
    return local_file_path



from os import listdir
from os.path import isfile, join
from krixik.__base__ import library_base_dir

data_directory = library_base_dir + "/utilities/validators/data"
not_data = ["__pycache__", "utilities", "__init__.py", "read_config.py", "config.yml", "decorators.py", "video.py"]
available_data_types = [
    name.split(".")[0]
    for name in [f for f in listdir(data_directory) if isfile(join(data_directory, f))]
    if name not in not_data and len(name.split(".")[0]) > 0
]



import os
import inspect
from functools import wraps

from krixik.utilities.validators.system.base.version import version_checker
from krixik.utilities.validators.system.base.file_ids import file_ids_checker
from krixik.utilities.validators.system.base.file_ids import individual_file_id_checker
from krixik.utilities.validators.system.base.file_tags import file_tags_checker
from krixik.utilities.validators.system.base.symbolic_paths import (
    symbolic_directory_paths_checker,
)
from krixik.utilities.validators.system.base.symbolic_paths import (
    individual_symbolic_directory_path_checker,
)
from krixik.utilities.validators.system.base.request_id import (
    request_id_checker,
)
from krixik.utilities.validators.system.base.verbose import verbose_checker
from krixik.utilities.validators.system.base.max_files import max_files_checker
from krixik.utilities.validators.system.base.file_description import (
    file_description_checker,
)
from krixik.utilities.validators.system.base.timestamp_bookends import (
    timestamp_bookends_checker,
)
from krixik.utilities.validators.system.base.local_save_directory import (
    local_save_directory_checker,
)
from krixik.utilities.validators.system.base.wait_for_process import (
    wait_for_process_checker,
)
from krixik.utilities.validators.system.base.sort_order import (
    sort_order_checker,
)
from krixik.utilities.validators.system.base.expire_time import expire_time_checker
from krixik.utilities.validators.system.base.query import query_checker
from krixik.utilities.validators.system.base.k_nn import k_checker
from krixik.utilities.validators.system.base.user_secrets import (
    api_key_checker,
    api_url_checker,
)
from krixik.utilities.validators.system.base.download_output import download_output_checker

from krixik.utilities.utilities import get_input
from typing import Callable


def type_check_inputs(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            signature = inspect.signature(func)

            verbose = get_input("verbose", signature, kwargs, default_value=True)
            verbose_checker(verbose)

            if "api_key" in list(kwargs.keys()):
                api_key_checker(kwargs["api_key"])

            if "api_url" in list(kwargs.keys()):
                api_url_checker(kwargs["api_url"])

            if "expire_time" in list(kwargs.keys()):
                expire_time_checker(kwargs["expire_time"])

            if "version" in list(kwargs.keys()):
                version_checker(kwargs["version"])

            if "file_ids" in list(kwargs.keys()):
                file_ids_checker(kwargs["file_ids"])

            if "file_id" in list(kwargs.keys()):
                individual_file_id_checker(kwargs["file_id"])

            if "file_tags" in list(kwargs.keys()):
                file_tags_checker(kwargs["file_tags"])

            if "symbolic_directory_paths" in list(kwargs.keys()):
                symbolic_directory_paths_checker(kwargs["symbolic_directory_paths"])

            if "symbolic_directory_path" in list(kwargs.keys()):
                individual_symbolic_directory_path_checker(kwargs["symbolic_directory_path"])

            sort_order = get_input("sort_order", signature, kwargs, default_value="descending")
            if sort_order is not None:
                sort_order_checker(sort_order)

            if "process_id" in list(kwargs.keys()):
                request_id_checker(kwargs["process_id"])

            if "request_id" in list(kwargs.keys()):
                request_id_checker(kwargs["request_id"])

            if "local_save_directory" in list(kwargs.keys()):
                local_save_directory_checker(kwargs["local_save_directory"])

            if "file_description" in list(kwargs.keys()):
                file_description_checker(kwargs["file_description"])

            download_output = get_input("download_output", signature, kwargs, default_value=True)
            download_output_checker(download_output)

            max_files = get_input("max_files", signature, kwargs)
            if max_files is not None:
                max_files_checker(max_files)

            wait_for_process = get_input("wait_for_process", signature, kwargs, default_value=False)
            wait_for_process_checker(wait_for_process)

            query = get_input("query", signature, kwargs)
            if query is not None:
                query_checker(query)

            k = get_input("k", signature, kwargs)
            if k is not None:
                k_checker(k)

            created_at_start = get_input("created_at_start", signature, kwargs)
            created_at_end = get_input("created_at_end", signature, kwargs)
            last_updated_start = get_input("last_updated_start", signature, kwargs)
            last_updated_end = get_input("last_updated_end", signature, kwargs)

            if created_at_start is not None or created_at_end is not None or last_updated_start is not None or last_updated_end is not None:
                timestamp_bookends_checker(
                    created_at_start,
                    created_at_end,
                    last_updated_start,
                    last_updated_end,
                    verbose=verbose,
                )

        except ValueError as e:
            raise ValueError(e)
        except TypeError as e:
            raise TypeError(e)
        except FileNotFoundError as e:
            raise FileNotFoundError(e)
        except PermissionError as e:
            raise PermissionError(e)
        except Exception as e:
            raise Exception(e)

        return func(*args, **kwargs)

    setattr(wrapper, "system_base_type_checked", bool(True))

    return wrapper


def file_converters_input_check(func):
    def wrapper(*args, **kwargs):
        if "local_file_path" in kwargs:
            if kwargs["local_file_path"] is not None:
                if os.path.exists(kwargs["local_file_path"]) is False:
                    raise FileNotFoundError(f"the file '{kwargs['local_file_path']}' does not exist.")
            else:
                raise ValueError("local_file_path cannot be None")
        if "local_save_directory" in kwargs:
            if kwargs["local_save_directory"] is not None:
                if os.path.exists(kwargs["local_save_directory"]) is False:
                    raise FileNotFoundError(f"the directory '{kwargs['local_save_directory']}' does not exist.")

        if "verbose" in kwargs:
            if kwargs["verbose"] is not None:
                if not isinstance(kwargs["verbose"], bool):
                    raise ValueError("verbose is not a boolean")
            else:
                raise ValueError("verbose cannot be None")
        return func(*args, **kwargs)

    return wrapper



from functools import wraps

from krixik.utilities.validators.data.audio import is_size as is_audio_size

# from krixik.utilities.validators.data.video import is_size as is_video_size
from krixik.utilities.validators.data.text import is_size as is_text_size
from krixik.utilities.validators.data.pdf import is_size as is_pdf_size
from krixik.utilities.validators.data.json import is_size as is_json_size
from krixik.utilities.validators.data.image import is_size as is_image_size
from krixik.utilities.validators.data.docx import is_size as is_docx_size
from krixik.utilities.validators.data.pptx import is_size as is_pptx_size
from krixik.utilities.validators.data.npy import is_size as is_npy_size

from krixik.utilities.validators.data.utilities.read_config import get_all_allowable_extensions


def datatype_validator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            if "local_file_path" in list(kwargs.keys()):
                local_file_path = kwargs["local_file_path"]
                extension = local_file_path.split(".")[-1]
                allowable_extensions = get_all_allowable_extensions()
                if "." + extension not in allowable_extensions:
                    raise ValueError(f"invalid file extension: '{extension}' - currently only {allowable_extensions} are allowed.")

                if extension == "txt":
                    is_text_size(local_file_path=local_file_path)
                elif extension == "pdf":
                    is_pdf_size(local_file_path=local_file_path)
                elif extension == "json":
                    is_json_size(local_file_path=local_file_path)
                elif extension == "jpg" or extension == "jpeg" or extension == "png":
                    is_image_size(local_file_path=local_file_path)
                elif extension == "docx":
                    is_docx_size(local_file_path=local_file_path)
                elif extension == "pptx":
                    is_pptx_size(local_file_path=local_file_path)
                elif extension == "mp3":
                    is_audio_size(local_file_path=local_file_path)
                # elif extension == "mp4":
                #     is_video_size(local_file_path=local_file_path)
                elif extension == "npy":
                    is_npy_size(local_file_path=local_file_path)
                else:
                    raise ValueError(f"invalid file extension: '{extension}'")
            return func(*args, **kwargs)
        except ValueError as e:
            raise ValueError(e)
        except Exception as e:
            raise Exception(e)

    return wrapper



import re

# define pattern for valid file id
id_pattern = r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$"


def is_valid_uuid_string(input_string: str) -> bool:
    return bool(re.match(id_pattern, input_string))


def request_id_checker(request_id: str) -> None:
    if request_id is not None:
        # check that request_id is a string
        if not isinstance(request_id, str):
            if request_id is None:
                raise TypeError("request_id must be input as an argument")
            else:
                raise TypeError(f"Invalid request_id: not a string -  {request_id}")

        # check that request_id matches pattern
        res = is_valid_uuid_string(request_id)
        if res is False:
            raise ValueError(f"Invalid _id: a valid _id is a string of length 36 with the following pattern: {id_pattern}")

        # check that request_id length is 36
        if len(request_id) != 36:
            raise ValueError(f"Invalid _id: length of this _id is not 36 (it must be exactly 36 characters) - {request_id}")



import re
from krixik.utilities.validators.system import (
    TAGS_COUNT_MAX,
    TAG_KEY_LENGTH_MAX,
    TAG_VALUE_LENGTH_MAX,
)

# validity pattern for file tags key and value
tags_pattern = r"^[a-zA-Z0-9\s\-_]*$"


def contains_only_alphanumeric_and_alphanumeric(input_string: str) -> bool:
    return bool(re.match(tags_pattern, input_string))


# check individual file tag
def individual_file_tag_checker(file_tag: dict, phase: str = "other") -> None:
    # check that file_tag is a dict
    if not isinstance(file_tag, dict):
        raise TypeError(f"file_tag must be input as a dictionary - {file_tag}")

    # check that file_tag is a dictionary with only one key-value pair
    if len(file_tag) != 1:
        raise ValueError(f"file_tag must be a dictionary with only one key-value pair - {file_tag}")

    # check key and value of file_tag to be valid strings
    key = list(file_tag.keys())[0]
    value = file_tag[key]

    # check that key is a string
    if not isinstance(key, str):
        raise TypeError(f"invalid file_tag key: not a string - {key}")

    # check that key length is greater than 0
    if len(key) == 0:
        raise ValueError(f"invalid file_tag key: length is 0 - {key}")

    # check that key length is less than 32
    if len(key) > TAG_KEY_LENGTH_MAX:
        raise ValueError(f"invalid file_tag key: length is greater than {TAG_KEY_LENGTH_MAX} (current maximum length allowable) - {key}")

    # check that key contains only alphanumeric and alphanumeric characters
    res = contains_only_alphanumeric_and_alphanumeric(key)
    if res is False:
        raise ValueError(f"invalid file_tag key: {key} - please check that it contains only alphanumeric and alphanumeric characters")

    # check that value is a string
    if not isinstance(value, str):
        raise ValueError(f"invalid file_tag value: not a string - {value}")

    # check that value length is greater than 0
    if len(value) == 0:
        raise ValueError(f"invalid file_tag value: length is 0 - {value}")

    # check that value length is less than 32
    if len(value) > TAG_VALUE_LENGTH_MAX:
        raise ValueError(f"invalid file_tag value: length is greater than {TAG_VALUE_LENGTH_MAX} (current maximum length allowable) - {value}")

    # if phase is process do not allow wildcard
    if phase == "process":
        if value == "*":
            raise ValueError(
                f"invalid file_tag value: {value} - please check that it is either a wildcard (*) or contains only alphanumeric and alphanumeric characters"
            )

    # check that value contains only alphanumeric and alphanumeric characters
    if value != "*":
        res = contains_only_alphanumeric_and_alphanumeric(value)
        if res is False:
            raise ValueError(f"invalid file_tag value: {value} - please check that it contains only alphanumeric and alphanumeric characters")


def lower_case_file_tags(file_tags: list) -> list:
    # lower case all tags
    lower_case_tags = []
    for file_tag in file_tags:
        key = list(file_tag.keys())[0]
        value = file_tag[key]
        lower_file_tag = {key.lower(): value.lower()}
        lower_case_tags.append(lower_file_tag)
    return lower_case_tags


def file_tags_checker(file_tags: list, phase: str = "other") -> None:
    if file_tags is not None:
        # check that file_tags is a list
        if not isinstance(file_tags, list):
            raise TypeError("file_tags is not a list")

        # check that file_tags is not empty
        lower_case_tags = []
        if file_tags is not None:
            # loop over each tag in file_tags and check for validity
            tag_count = 0
            for file_tag in file_tags:
                # check that tag_count is less or equal to 20
                if tag_count > TAGS_COUNT_MAX - 1:
                    raise ValueError(
                        f"Invalid file_tags: {file_tags} - more than {TAGS_COUNT_MAX} tags provided, you can have a maximum of {TAGS_COUNT_MAX} tags"
                    )

                # type check each file tag
                individual_file_tag_checker(file_tag, phase=phase)

                # increment tag_count
                tag_count += 1

            # lower case tags
            lower_case_tags = lower_case_file_tags(file_tags)

            # check for duplicate dicts in lower_case_file_tags
            if len(lower_case_tags) != len(set(map(lambda x: frozenset(x.items()), lower_case_tags))):
                # determine duplicates
                duplicates = []
                seen = set()
                for tag in lower_case_tags:
                    if frozenset(tag.items()) in seen:
                        duplicates.append(tag)
                    else:
                        seen.add(frozenset(tag.items()))
                raise ValueError(f"file_tags contains duplicate key-value pairs: {duplicates}")
        else:
            raise ValueError("file_tags is None")



import os
from typing import Optional
from krixik.utilities.validators.data.text import is_size
from krixik.utilities.validators.system.base.utilities.decorators import (
    file_converters_input_check,
)
from krixik.utilities.converters.default_clean_options import (
    default_clean_options,
    default_clean_keys,
)
from krixik.utilities.utilities import vprint


def delete_file(*, file_path: str, exception: str) -> None:
    if file_path is not None:
        # remove file_path file if it exists
        if os.path.exists(file_path):
            os.remove(file_path)
        raise ValueError(f"there seems to be a problem with your cleaned file - file cleaning failed with exception {str(exception)}")


@file_converters_input_check
def convert(
    *,
    local_file_path: Optional[str],
    local_save_directory: Optional[str],
    clean_options: Optional[dict] = None,
    use_default_clean_options: bool = False,
    verbose: bool = True,
) -> Optional[str]:
    """cleans input text file, replacing un-processable characters

    Parameters
    ----------
    local_file_path : str | None
        path to local file to convert
    local_save_directory : str | None
        local directory to save converted file
    clean_options: dict | None
        user-defined dictionary of character replacements
    use_default_clean_options: bool
        flag to use default system replacement options, defaults to False
    verbose : bool, optional
        by default True

    Returns
    -------
    str | None
        if cleaning successful returns path to cleaned file
    """

    if local_file_path is None:
        raise NameError("local_file_path not defined")
    if os.path.exists(local_file_path) is False:
        raise FileNotFoundError(f"the file '{local_file_path}' does not exist.")
    if local_save_directory is None:
        raise NameError("local_save_directory is not defined")

    new_file_path = ""
    try:
        # new file to write to
        file_name = "krixik_cleaned_version_" + os.path.splitext(os.path.basename(local_file_path))[0]
        extension = ".txt"
        new_file_path = os.path.join(local_save_directory, file_name + extension)

        # remove file if it already exists
        if os.path.exists(new_file_path):
            os.remove(new_file_path)

        # check that input file is a valid text file and sized correctly
        is_size(local_file_path=local_file_path)

        if clean_options is not None:
            keys = set(list(clean_options.keys()))
            values = set(list(clean_options.values()))
            intersection = keys.intersection(values)
            if len(intersection) > 0:
                raise ValueError(f"clean_options contains the same key and value: {intersection}")

        # if clean_options length greater than 0 convert to one dict
        clean_keys = []

        if clean_options is not None:
            if len(clean_options) > 0:
                clean_keys = list(clean_options.keys())
                if use_default_clean_options is True:
                    vprint(
                        "WARNING: You are using both default_clean_options and your custom clean_options.  Note: any duplication in your clean_options will overwrite the same key in default_clean_options.",
                        verbose=verbose,
                    )

        # clean file
        clean_count = 0
        line_count = 0
        with open(local_file_path, "r", encoding="utf-8") as file:
            for line in file:
                new_line = ""
                for char in line:
                    new_char = char
                    # replace char with clean_options[char] if char in clean_options
                    if clean_options is not None:
                        if len(clean_options) > 0:
                            if char in clean_keys:
                                new_char = clean_options[char]
                                clean_count += 1

                    # replace char with default_clean_options[char] if char in default_clean_options
                    if use_default_clean_options is True:
                        if char in default_clean_keys:
                            new_char = default_clean_options[char]
                            clean_count += 1

                    new_line += new_char

                # write line to new file
                if line_count == 0:
                    with open(new_file_path, "a", encoding="utf-8") as new_file:
                        new_file.write(new_line)

                if line_count > 0:
                    if os.path.exists(new_file_path):
                        with open(new_file_path, "a", encoding="utf-8") as new_file:
                            new_file.write(new_line)
                    else:
                        raise FileNotFoundError(f"file {new_file_path} does not exist")

                # check size to make sure file is not oversize - every 50 lines
                if line_count % 50 == 0 and line_count > 0:
                    is_size(local_file_path=new_file_path)

                # update counter
                line_count += 1

        if clean_count == 0:
            vprint(
                f"INFO: No characters needed cleaning from the original file: {local_file_path}",
                verbose=verbose,
            )
            vprint(
                f"INFO: Removing newly created (redundant) clean file: {new_file_path}",
                verbose=verbose,
            )
            os.remove(new_file_path)
            return local_file_path
        else:
            vprint(
                f"SUCCESS: File cleaning complete, result saved to: {new_file_path}",
                verbose=verbose,
            )

            vprint(
                "INFO: Checking that file size falls within acceptable parameters...",
                verbose=verbose,
            )
            is_size(local_file_path=new_file_path)
            vprint("INFO:...success!", verbose=verbose)
            return new_file_path
    except ValueError as ve:
        vprint(
            "INFO:...failure!  The converted (text) file does not fall within acceptable size parameters.",
            verbose=verbose,
        )
        delete_file(file_path=new_file_path, exception=str(ve))
    except Exception as e:
        raise ValueError(f"file cleaning failed with exception: {e}")



import requests
import json
import datetime
from typing import Optional
from krixik.utilities.utilities import vprint
from krixik.utilities.validators.system.base.timestamp_bookends import (
    convert_timestamps,
)
from krixik.system_builder.functions import list_endpoint


# method for listing files
def list_files(
    self,
    *,
    file_ids: Optional[list] = None,
    file_names: Optional[list] = None,
    symbolic_directory_paths: Optional[list] = None,
    symbolic_file_paths: Optional[list] = None,
    file_tags: Optional[list] = None,
    sort_order: str = "descending",
    max_files: Optional[int] = None,
    created_at_start: Optional[str] = None,
    created_at_end: Optional[str] = None,
    last_updated_start: Optional[str] = None,
    last_updated_end: Optional[str] = None,
    verbose: bool = True,
) -> dict:
    # check that at least one query argument is given
    if (
        file_ids is None
        and file_names is None
        and symbolic_directory_paths is None
        and symbolic_file_paths is None
        and file_tags is None
        and created_at_start is None
        and created_at_end is None
        and last_updated_start is None
        and last_updated_end is None
    ):
        raise ValueError("please provide at least one query argument")

    # if max_files is None, set to default value of 1000
    if max_files is None:
        max_files = 1000
        vprint(
            "INFO: value of max_files not set by user, setting max_files to default value of 1000.",
            verbose=verbose,
        )

    if sort_order is None:
        sort_order = "descending"
        vprint(
            "INFO: value of sort_order not set by user, setting sort_order to default value of descending.",
            verbose=verbose,
        )
    else:
        vprint(f"INFO: sort_order set to {sort_order}", verbose=verbose)

    # check timestamp bookends
    (
        created_at_start_int,
        created_at_end_int,
        last_updated_start_int,
        last_updated_end_int,
    ) = convert_timestamps(created_at_start, created_at_end, last_updated_start, last_updated_end)

    if hasattr(self, "_KrixikBasePipeline__pipeline"):
        pipeline = self._KrixikBasePipeline__pipeline
    elif hasattr(self, "_KrixikSearchPipeline__pipeline"):
        pipeline = self._KrixikSearchPipeline__pipeline
    else:
        raise ValueError("pipeline not found in self")

    if hasattr(self, "_KrixikBasePipeline__version"):
        version = self._KrixikBasePipeline__version
    elif hasattr(self, "_KrixikSearchPipeline__version"):
        version = self._KrixikSearchPipeline__version
    else:
        raise ValueError("version not found in self")

    # prep payload data
    payload_data = {
        "pipeline": pipeline,
        "version": version,
        "file_ids": file_ids,
        "file_names": file_names,
        "symbolic_directory_paths": symbolic_directory_paths,
        "symbolic_file_paths": symbolic_file_paths,
        "file_tags": file_tags,
        "max_files": max_files,
        "created_at_start": created_at_start_int,
        "created_at_end": created_at_end_int,
        "last_updated_start": last_updated_start_int,
        "last_updated_end": last_updated_end_int,
    }

    if hasattr(self, "_KrixikBasePipeline__api_key") and hasattr(self, "_KrixikBasePipeline__api_url"):
        api_key = self._KrixikBasePipeline__api_key
        api_url = self._KrixikBasePipeline__api_url
    elif hasattr(self, "_KrixikSearchPipeline__api_key") and hasattr(self, "_KrixikSearchPipeline__api_url"):
        api_key = self._KrixikSearchPipeline__api_key
        api_url = self._KrixikSearchPipeline__api_url
    else:
        raise ValueError("api_key and api_url not found in self")

    # prep headers
    headers = {"Content-Type": "text/plain", "krixikApiKey": api_key}

    try:
        # make request
        response = requests.post(api_url + list_endpoint, headers=headers, json=payload_data, timeout=60)

        # return response
        results = json.loads(response.text)
        if "request_id" in results:
            self.request_id = results["request_id"]
        status_code_dict = {"status_code": response.status_code}
        if response.status_code == 200:
            if sort_order == "descending":
                # sort user_items based on created_at
                results["items"] = sorted(
                    results["items"],
                    key=lambda x: datetime.datetime.strptime(x["created_at"], "%Y-%m-%d %H:%M:%S").timestamp(),
                    reverse=True,
                )
            elif sort_order == "ascending":
                # sort user_items based on created_at
                results["items"] = sorted(
                    results["items"],
                    key=lambda x: datetime.datetime.strptime(x["created_at"], "%Y-%m-%d %H:%M:%S").timestamp(),
                    reverse=False,
                )
        status_code_dict.update(results)
        return status_code_dict
    except Exception as e:
        raise ValueError(f"list failed with request exception {e}")
    finally:
        self._reset_class_variables()



from dataclasses import dataclass, field
from typing import Literal, Dict


@dataclass
class InputStructure:
    format: Literal["audio"] = "audio"
    filename: str = "input_audio.mp3"
    process_key: None = None

    @property
    def data_example(self):
        return None

    @property
    def process_type(self):
        return str(self.__annotations__[self.process_key]) if self.process_key is not None else None


@dataclass
class OutputStructure:
    format: Literal["json"] = "json"
    filename: str = "filename_example.json"
    process_key: str = "transcript"
    transcript: str = "This is the full transcript."

    segments: list = field(
        default_factory=lambda: [
            {
                "id": 1,
                "seek": 0,
                "start": 0.0,
                "end": 10.0,
                "text": "This is the",
                "tokens": [20, 34],
                "temperature": 0.0,
                "avg_logprob": 0.0,
                "compression_ratio": 0.0,
                "no_speech_prob": 0.0,
                "confidence": 0.0,
                "words": [
                    {"text": "This", "start": 0.0, "end": 1.0, "confidence": 0.5},
                    {"text": "is the", "start": 1.0, "end": 2.0, "confidence": 0.6},
                ],
            },
            {
                "id": 2,
                "seek": 10,
                "start": 10.0,
                "end": 20.0,
                "text": "main text",
                "tokens": [44, 101],
                "temperature": 0.0,
                "avg_logprob": 0.0,
                "compression_ratio": 0.0,
                "no_speech_prob": 0.0,
                "confidence": 0.0,
                "words": [
                    {"text": "main", "start": 10.0, "end": 11.0, "confidence": 0.7},
                    {"text": "text", "start": 11.0, "end": 12.0, "confidence": 0.8},
                ],
            },
        ]
    )
    language: Dict[str, str] = field(default_factory=lambda: {"language": "English"})

    @property
    def process_type(self):
        return str(self.__annotations__[self.process_key]) if self.process_key is not None else None

    @property
    def data_example(self):
        return {
            "transcript": self.transcript,
            "segments": self.segments,
            "language": self.language["language"],
        }



def process_switches_checker(*, process_for_vector_search: bool, process_for_keyword_search: bool) -> None:
    if process_for_vector_search is not None and process_for_keyword_search is not None:
        # check processing switches
        if not isinstance(process_for_vector_search, bool):
            raise TypeError("process_for_vector_search must be a bool")

        if not isinstance(process_for_keyword_search, bool):
            raise TypeError("process_for_keyword_search must be a bool")

        if process_for_vector_search is False and process_for_keyword_search is False:
            raise ValueError("process_for_vector_search and process_for_keyword_search cannot both be False")



import os


def local_file_path_checker(local_file_path: str) -> None:
    if local_file_path is None:
        raise ValueError("local_file_path cannot be empty")

    if isinstance(local_file_path, str) is False:
        raise TypeError(f"invalid local file path: not a string - {local_file_path}")

    if os.path.exists(local_file_path) is False:
        raise FileNotFoundError(f"invalid local file path: does not exist - {local_file_path}")



import os
from moviepy.editor import VideoFileClip
from pydub.exceptions import PydubException


def is_valid(local_file_path: str) -> None:
    if not os.path.exists(local_file_path):
        raise ValueError(f"input file does not exist - {local_file_path}")

    if not local_file_path.endswith(".mp4"):
        raise ValueError(f"input file must end with .mp4 extension, not a valid video file - {local_file_path}")

    try:
        start_time = 0
        end_time = 3
        clip = VideoFileClip(local_file_path).subclip(start_time, end_time)
        clip.close()
    except PydubException:
        raise ValueError(f"input file is not a valid video file - {local_file_path}")
    except FileNotFoundError:
        raise FileNotFoundError(f"The file '{local_file_path}' does not exist.")
    except Exception as e:
        raise ValueError(f"video validation failed with exception {e}")


def is_size(
    *,
    local_file_path: str,
    minimum_seconds: int = 1,
    maximum_seconds: int = 180,
    minimum_file_size: float = 0.2,
    maximum_file_size: float = 100.000001,
) -> None:
    # proper size
    def compute_size(file_path: str):
        try:
            # Get the size of the file in bytes
            file_size_bytes = os.path.getsize(file_path)

            # Convert the size to megabytes (MB)
            file_size_mb = file_size_bytes / (1024 * 1024)
            return file_size_mb
        except Exception as e:
            raise ValueError(f"audio compute size failed with exception {e}")

    def compute_duration(file_path):
        try:
            video_clip = VideoFileClip(file_path)
            duration_in_seconds = round(video_clip.duration, 1)
            return duration_in_seconds
        except Exception as e:
            raise ValueError(f"audio compute duration failed with exception {e}")

    # check that local_file_path is valid
    is_valid(local_file_path)

    try:
        # compute file size
        file_size = compute_size(local_file_path)

        # check that file size in megabytes is greater than minimum_file_size and less than maximum_file_size
        if file_size < minimum_file_size:
            raise ValueError(
                f"input file size is {round(file_size,2)} megabytes: less than {minimum_file_size} megabytes (current minimum size allowable)"
            )

        if file_size > maximum_file_size:
            raise ValueError(
                f"*** Krixik Open Beta warning *** input file size is {round(file_size,2)} megabytes: greater than {maximum_file_size} megabytes (current maximum size allowable) - {local_file_path} during Krixik Open Beta"
            )

        # compute length of audio file in seconds
        audio_duration = compute_duration(local_file_path)

        # check that audio file is  greater than minimum_seconds and less than maximum_seconds
        if audio_duration < minimum_seconds:
            raise ValueError(f"audio file is {audio_duration} seconds - this is less than {minimum_seconds} seconds (current minimum size allowable)")
        if audio_duration > maximum_seconds:
            raise ValueError(f"audio file or greater than {maximum_seconds} seconds (current maximum size allowable)")

    except ValueError as ve:
        raise ve
    except Exception as e:
        raise ValueError(f"audio extraction failed with exception {e}")



from typing import Dict, List


def convert_paths_to_dict(paths: List[str]) -> Dict[str, str]:
    root: Dict[str, str] = {}
    for path in paths:
        # separate by slashes, disregarding the first `/`
        path = path.lstrip("/").split("/")
        # pop off the last key-value component
        key, _, val = path.pop(-1).partition("=")
        # find the target dict starting from the root
        target_dict = root
        for component in path:
            target_dict: Dict[str, str] = target_dict.setdefault(component, {})
        # assign key-value
        target_dict[key] = val
    return root


def tree(paths: List[str], prefix: str = "") -> str:
    """
    A recursive generator, given a directory Path object
    will yield a visual tree structure line by line
    with each line prefixed by the same characters
    """

    # prefix components:
    space = "    "
    branch = "│   "
    # pointers
    tee = "├── "
    last = "└── "

    # contents each get pointers that are ├── with a final └── :
    pointers = [tee] * (len(paths) - 1) + [last]
    for pointer, path in zip(pointers, paths):
        if "." not in path:
            yield prefix + pointer + "/" + path
        else:
            yield prefix + pointer + path
        if isinstance(paths[path], dict):  # extend the prefix and recurse:
            extension = branch if pointer == tee else space
            # i.e. space because last, └── , above so no more │
            yield from tree(paths[path], prefix=prefix + extension)


def show_symbolic_file_tree(paths: List[str]) -> int:
    try:
        # convert paths to dict
        dict_paths = convert_paths_to_dict(paths)

        # Display root separately
        print("/")

        # Print tree of dict_paths
        for line in tree(dict_paths):
            print(line)

        return 0
    except Exception as e:
        print(f"FAILURE: show tree illustrator failed with exception {e}")
        return 1



def name_check(name: str) -> None:
    if not isinstance(name, str):
        raise ValueError(f"FAILURE: your custom pipeline name - {name} - is not a string")
    if len(name) == 0 or len(name) > 64:
        raise ValueError(f"FAILURE: your name - {name} - must be greater than 0 and less than 64 characters")



from dataclasses import dataclass, field
from typing import Literal
import numpy as np


@dataclass
class InputStructure:
    format: Literal["npy"] = "npy"
    filename: str = "output_data.npy"
    process_key: str = "data"
    data: np.ndarray = field(default_factory=lambda: np.array([1, 2, 3]))

    @property
    def process_type(self):
        return str(self.__annotations__[self.process_key]) if self.process_key is not None else None

    @property
    def data_example(self):
        return self.data


@dataclass
class OutputStructure:
    format: Literal["faiss"] = "faiss"
    filename: str = "output_index.faiss"
    process_key: None = None

    @property
    def process_type(self):
        return str(self.__annotations__[self.process_key]) if self.process_key is not None else None

    @property
    def data_example(self):
        return "this will be a database"



from krixik.utilities.validators.system import MAX_FILES_COUNT


def max_files_checker(max_files: int) -> None:
    if max_files is not None:
        # check that max_files is an integer
        if not isinstance(max_files, int) or isinstance(max_files, bool):
            raise TypeError(f"invalid max_files: not an integer - {max_files}")

        if max_files < 1:
            raise ValueError(f"invalid max_files: less than 1 - {max_files}")

        if max_files > MAX_FILES_COUNT:
            raise ValueError(f"invalid max_files: greater than {MAX_FILES_COUNT} - {max_files}")

        if max_files < 1:
            raise ValueError(f"invalid max_files: less than 1 - {max_files}")



import json
import requests
from datetime import datetime
from typing import List, Optional

from krixik.system_builder.functions import keyword_search_endpoint
from krixik.utilities.utilities import vprint
from krixik.utilities.validators.system.base.timestamp_bookends import (
    convert_timestamps,
)
from krixik.system_builder.functions.checkin import check_init_decorator
from krixik.utilities.validators.utilities.decorators import type_check_inputs
from krixik.system_builder.utilities.decorators import kwargs_checker


@kwargs_checker
@check_init_decorator
@type_check_inputs
def keyword_search(
    self,
    *,
    query: Optional[str] = None,
    file_ids: Optional[list] = None,
    file_names: Optional[List] = None,
    symbolic_directory_paths: Optional[List] = None,
    symbolic_file_paths: Optional[List] = None,
    file_tags: Optional[dict] = None,
    sort_order: Optional[str] = "descending",
    max_files: Optional[int] = None,
    created_at_start: Optional[str] = None,
    created_at_end: Optional[str] = None,
    last_updated_start: Optional[str] = None,
    last_updated_end: Optional[str] = None,
    verbose: bool = False,
) -> dict:
    """keyword search over files defined by query parameters (e.g., file_ids, file_names, etc.)

    Parameters
    ----------
    query : str | None, optional
        query to search for, a string that can consist of multiple words, by default None
    file_ids : List | None, optional
        list of file_ids' files to include in the search, by default None
    file_names : List | None, optional
        list of file_names' files to include in the search, by default None
    symbolic_directory_paths : List | None, optional
        list of symbolic_directory_paths' or stumps' files to include in the search, by default None
    symbolic_file_paths : List | None, optional
        list of symbolic_file_paths' files to include in the search, by default None
    file_tags : dict | None, optional
        list of file_tags' or file_tag stumps' files to include in the search, by default None
    sort_order : str, optional
        the sort order of the files, by default "descending"
    max_files : int | None, optional
        the maximum number of files to return, by default None
    created_at_start : str | None, optional
        start date of files to include in the search, by default None
    created_at_end : str | None, optional
        end date of files to include in the search, by default None
    last_updated_start : str | None, optional
        start date of files to include in the search, by default None
    last_updated_end : str | None, optional
        end date of files to include in the search, by default None
    verbose : bool, optional
        whether to print verbose output, by default True

    Returns
    -------
    dict
        a dictionary containing the status_code, request_id, and message indicating success or failure
    """
    if sort_order == "global":
        raise ValueError("sort order must be either 'descending' or 'ascending'")

    # check if query is empty
    if query == "" or query is None:
        raise ValueError("query is empty - please enter a query")

    # check that not all query args are None
    if (
        file_ids is None
        and file_names is None
        and symbolic_directory_paths is None
        and symbolic_file_paths is None
        and file_tags is None
        and created_at_start is None
        and created_at_end is None
        and last_updated_start is None
        and last_updated_end is None
    ):
        raise ValueError("please provide at least one query argument")

    # if max_files is None, set to default value of 1000
    if max_files is None:
        max_files = 1000
        vprint(
            "INFO: value of max_files not set by user, setting max_files to default value of 1000.",
            verbose=verbose,
        )

    if sort_order is None:
        sort_order = "descending"
        vprint(
            "INFO: value of sort_order not set by user, setting sort_order to default value of descending.",
            verbose=verbose,
        )

    vprint(f"INFO: sort_order is set to {sort_order}.", verbose=verbose)

    # check timestamp bookends
    (
        created_at_start_int,
        created_at_end_int,
        last_updated_start_int,
        last_updated_end_int,
    ) = convert_timestamps(created_at_start, created_at_end, last_updated_start, last_updated_end)

    # prep payload data
    payload_data = {
        "pipeline": self._KrixikBasePipeline__pipeline,
        "pipeline_ordered_modules": self._KrixikBasePipeline__pipeline_ordered_modules,
        "query": query,
        "file_ids": file_ids,
        "file_names": file_names,
        "symbolic_directory_paths": symbolic_directory_paths,
        "symbolic_file_paths": symbolic_file_paths,
        "file_tags": file_tags,
        "max_files": max_files,
        "created_at_start": created_at_start_int,
        "created_at_end": created_at_end_int,
        "last_updated_start": last_updated_start_int,
        "last_updated_end": last_updated_end_int,
    }

    # prep headers
    headers = {"Content-Type": "text/plain", "krixikApiKey": self._KrixikBasePipeline__api_key}

    try:
        # make request
        response = requests.post(
            (self._KrixikBasePipeline__api_url or "") + keyword_search_endpoint,
            headers=headers,
            json=payload_data,
            timeout=60,
        )

        # return response
        results = json.loads(response.text)

        status_code_dict = {"status_code": response.status_code}
        if response.status_code == 200:
            if sort_order == "descending":
                # sort user_items based on created_at
                results["items"] = sorted(
                    results["items"],
                    key=lambda x: datetime.strptime(x["file_metadata"]["created_at"], "%Y-%m-%d %H:%M:%S").timestamp(),
                    reverse=True,
                )

            if sort_order == "ascending":
                # sort user_items based on created_at
                results["items"] = sorted(
                    results["items"],
                    key=lambda x: datetime.strptime(x["file_metadata"]["created_at"], "%Y-%m-%d %H:%M:%S").timestamp(),
                    reverse=False,
                )

        status_code_dict.update(results)
        return status_code_dict
    except Exception as e:
        raise ValueError(f"keyword search failed with request exception {e}") from e



import os
from pypdf import PdfReader


def is_valid(local_file_path: str) -> None:
    if local_file_path.split(".")[-1] != "pdf":
        raise ValueError(f"invalid local_file_path: must be a pdf file - {local_file_path}")

    try:
        # check that local_file_path represents pdf file
        reader = PdfReader(local_file_path)

        # all text
        for page in reader.pages:
            if page is not None:
                page.extract_text()
                return None
        raise ValueError(f"file is not a pdf file - {local_file_path}")
    except FileNotFoundError:
        raise FileNotFoundError(f"The file '{local_file_path}' does not exist.")
    except Exception as e:
        raise ValueError(f"pdf extraction failed with exception {e}")


def is_size(
    *,
    local_file_path: str,
    minimum_word_count: int = 10,
    minimum_file_size: float = 0.000001,
    maximum_file_size: float = 100.000001,
) -> None:
    # proper size
    def compute_size(file_path: str):
        try:
            # Get the size of the file in bytes
            file_size_bytes = os.path.getsize(file_path)

            # Convert the size to megabytes (MB)
            file_size_mb = file_size_bytes / (1024 * 1024)
            return file_size_mb
        except Exception as e:
            raise ValueError(f"pdf size calculation failed with exception {e}")

    # compute word count
    def compute_word_count(file_path: str):
        try:
            # check that local_file_path represents pdf file
            reader = PdfReader(file_path)

            # extract all text
            word_count = 0
            for page in reader.pages:
                if page is None:
                    pass
                else:
                    line = page.extract_text()
                    words = line.split()
                    word_count += len(words)
            return word_count
        except Exception as e:
            raise ValueError(f"pdf word count calculation failed with exception {e}")

    # check size of input text file
    try:
        # check that local_file_path represents a valid text file
        is_valid(local_file_path)

        # compute file size in megabytes
        file_size = compute_size(local_file_path)

        # check that file size in megabytes is greater than minimum_file_size and less than maximum_file_size
        if file_size < minimum_file_size:
            raise ValueError(
                f"input file size is {round(file_size,2)} megabytes: less than {minimum_file_size} megabytes (current minimum size allowable)"
            )

        if file_size > maximum_file_size:
            raise ValueError(
                f"*** Krixik Open Beta warning *** input file size is {round(file_size,2)} megabytes: greater than {maximum_file_size} megabytes (current maximum size allowable) - {local_file_path} during Krixik Open Beta"
            )

        # compute word count
        file_word_count = compute_word_count(local_file_path)

        # check that word count is greater than minimum_word_count
        if file_word_count < minimum_word_count:
            raise ValueError(f"it appears that the word count is less than the allowed minimum {minimum_word_count} words")

    except ValueError as ve:
        raise ve

    except Exception as e:
        raise ValueError(f"pdf extraction failed with exception {e}")



# import os
# from moviepy.editor import VideoFileClip
# from krixik.utilities.validators.system.base.utilities.decorators import (
#     file_converters_input_check,
# )
# from krixik.utilities.validators.data.video import is_size as is_video_size
# from krixik.utilities.validators.data.audio import is_size as is_audio_size
# from krixik.utilities.utilities import vprint


# def delete_file(*, file_path: str, exception: str) -> None:
#     if file_path is not None:
#         if os.path.exists(file_path):
#             os.remove(file_path)
#         raise ValueError(f"audio extraction failed with exception: {str(exception)}")


# @file_converters_input_check
# def extract_audio(*, local_file_path: str, audio_filepath: str) -> None:
#     try:
#         video = VideoFileClip(local_file_path)
#         audio = video.audio
#         if audio is not None:
#             audio.write_audiofile(audio_filepath, verbose=False, logger=None)
#             audio.close()
#             video.close()
#     except Exception as e:
#         raise ValueError(f"error extracting audio from video {local_file_path}, exception: {e}")


# @file_converters_input_check
# def convert(
#     *,
#     local_file_path: str | None,
#     local_save_directory: str | None,
#     verbose: bool = True,
# ) -> str:
#     """video to audio converter - strips audio mp3 from video file using moviepy library

#     Parameters
#     ----------
#     local_file_path : str | None
#         path to local file to convert
#     local_save_directory : str | None
#         local directory to save converted file
#     verbose : bool, optional
#         by default True

#     Returns
#     -------
#     str | None
#         if conversion successful returns path to converted file
#     """

#     if local_file_path is not None and local_save_directory is not None:
#         audio_filepath = None
#         try:
#             # get file_name video_file_path
#             file_name = "krixik_converted_version_" + os.path.splitext(os.path.basename(local_file_path))[0]

#             # format audio file path
#             extension = ".mp3"
#             audio_filepath = os.path.join(local_save_directory, file_name + extension)

#             # remove file if it already exists
#             if os.path.exists(audio_filepath):
#                 os.remove(audio_filepath)

#             # check input file is mp4
#             is_video_size(local_file_path=local_file_path, minimum_file_size=0.1)

#             # extract audio from video
#             extract_audio(local_file_path=local_file_path, audio_filepath=audio_filepath)

#             # report size check
#             vprint(
#                 "INFO: Checking that file size falls within acceptable parameters...",
#                 verbose=verbose,
#             )
#             is_audio_size(local_file_path=audio_filepath, minimum_file_size=0.05)
#             vprint("INFO:...success!", verbose=verbose)

#             return audio_filepath
#         except ValueError as ve:
#             vprint(
#                 "INFO:...failure!  The converted (audio) file does not fall within acceptable size parameters.",
#                 verbose=verbose,
#             )
#             if audio_filepath is not None:
#                 delete_file(file_path=audio_filepath, exception=str(ve))
#             raise
#         except AttributeError as ae:
#             if audio_filepath is not None:
#                 delete_file(file_path=audio_filepath, exception=str(ae))
#             raise ValueError(f"video contains no visual content, cannot convert to audio - {local_file_path}")
#         except Exception as e:
#             if audio_filepath is not None:
#                 delete_file(file_path=audio_filepath, exception=str(e))
#             raise
#     else:
#         raise ValueError("the input local_file_path and/or local_save_directory is null")
